from typing import Any, Iterable

import numpy as np


def check_arrays(values: Iterable[Any]) -> None:
    """Checks whether paths to numpy arrays are valid. This checks that th file exists
    and is readable by numpy.

    @type values: List[Any]
    @param values: A list of paths to numpy arrays.
    @rtype: NoneType
    @return: None
    """

    def _check_valid_array(path: str) -> bool:
        try:
            np.load(path)
            return True
        except Exception:
            return False

    for value in values:
        if not isinstance(value, str):
            raise Exception(
                f"Array value {value} must be a path to a numpy array (.npy)"
            )
        if not _check_valid_array(value):
            raise Exception(f"Array at path {value} is not a valid numpy array (.npy)")



from .archive_generator import ArchiveGenerator
from .config import Config
from .model import Model
from .utils import is_nn_archive

__all__ = ["ArchiveGenerator", "Model", "Config", "is_nn_archive"]



import shutil
from pathlib import Path

import pytest

from luxonis_ml.utils import setup_logging
from luxonis_ml.utils.environ import environ

setup_logging(
    use_rich=True,
    rich_print=True,
    # level="DEBUG",
    configure_warnings=True,
)


@pytest.fixture(autouse=True, scope="module")
def set_paths():
    environ.LUXONISML_BASE_PATH = Path.cwd() / "tests/data/luxonisml_base_path"
    if environ.LUXONISML_BASE_PATH.exists():
        shutil.rmtree(environ.LUXONISML_BASE_PATH)

    environ.LUXONISML_TMP_DIR = Path.cwd() / "tests/data/luxonisml_tmp_dir"
    if environ.LUXONISML_TMP_DIR.exists():
        shutil.rmtree(environ.LUXONISML_TMP_DIR)



import hashlib
import json
import logging
from pathlib import Path
from typing import List, Optional, Tuple

import cv2
import numpy as np
import rich.box
import typer
import yaml
from rich import print
from rich.console import Console
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from typing_extensions import Annotated

from luxonis_ml.data import (
    Augmentations,
    LabelType,
    LuxonisDataset,
    LuxonisLoader,
    LuxonisParser,
)
from luxonis_ml.enums import DatasetType, SplitType

logger = logging.getLogger(__name__)

app = typer.Typer()


def complete_dataset_name(incomplete: str):
    datasets = LuxonisDataset.list_datasets()
    return [d for d in datasets if d.startswith(incomplete)]


DatasetNameArgument = Annotated[
    str,
    typer.Argument(
        ..., help="Name of the dataset.", autocompletion=complete_dataset_name
    ),
]


def check_exists(name: str):
    if not LuxonisDataset.exists(name):
        print(f"[red]Dataset [magenta]'{name}'[red] does not exist.")
        raise typer.Exit()


def get_dataset_info(name: str) -> Tuple[int, List[str], List[str]]:
    dataset = LuxonisDataset(name)
    try:
        size = len(dataset)
    except KeyError:
        size = -1

    try:
        loader = LuxonisLoader(dataset, view=SplitType.TRAIN.value)
        _, ann = next(iter(loader))
    except Exception:
        ann = {}
    classes, _ = dataset.get_classes()
    tasks = list(ann.keys())
    return size, classes, tasks


def print_info(name: str) -> None:
    size, classes, tasks = get_dataset_info(name)
    print(
        Panel.fit(
            f"[magenta b]Name: [not b cyan]{name}\n"
            f"[magenta b]Size: [not b cyan]{size}\n"
            f"[magenta b]Classes: [not b cyan]{', '.join(classes)}\n"
            f"[magenta b]Tasks: [not b cyan]{', '.join(tasks)}",
            title="Dataset Info",
        )
    )


@app.command()
def info(name: DatasetNameArgument):
    """Prints information about a dataset."""
    check_exists(name)
    print_info(name)


@app.command()
def delete(name: DatasetNameArgument):
    """Deletes a dataset."""
    check_exists(name)

    dataset = LuxonisDataset(name)
    dataset.delete_dataset()
    print(f"Dataset '{name}' deleted.")


@app.command()
def ls(
    full: Annotated[
        bool, typer.Option("--full", "-f", help="Show full information.")
    ] = False,
):
    """Lists all datasets."""
    datasets = LuxonisDataset.list_datasets()
    table = Table(
        title="Datasets" + (" - Full Table" if full else ""),
        box=rich.box.ROUNDED,
        row_styles=["yellow", "cyan"],
    )
    table.add_column("Name", header_style="magenta i")
    table.add_column("Size", header_style="magenta i")
    if full:
        table.add_column("Classes", header_style="magenta i")
        table.add_column("Tasks", header_style="magenta i")
    for name in datasets:
        dataset = LuxonisDataset(name)
        rows = [name]
        try:
            size = len(dataset)
        except KeyError:
            size = -1
        rows.append(str(size))
        if full:
            _, classes, tasks = get_dataset_info(name)
            rows.extend(
                [
                    ", ".join(classes) if classes else "[red]<empty>[no red]",
                    ", ".join(tasks) if tasks else "[red]<empty>[no red]",
                ]
            )
        table.add_row(*rows)
    console = Console()
    console.print(table)


@app.command()
def inspect(
    name: DatasetNameArgument,
    view: Annotated[
        SplitType,
        typer.Option(
            ...,
            "--view",
            "-v",
            help="Which split of the dataset to inspect.",
            case_sensitive=False,
        ),
    ] = "train",  # type: ignore
    aug_config: Annotated[
        Optional[Path],
        typer.Option(
            ...,
            "--aug-config",
            "-a",
            help="Path to a config defining augmentations. "
            "This can be either a json or a yaml file.",
        ),
    ] = None,
):
    """Inspects images and annotations in a dataset."""

    def _task_to_rgb(string: str) -> tuple:
        h = int(hashlib.md5(string.encode()).hexdigest(), 16)
        r = (h & 0xFF0000) >> 16
        g = (h & 0x00FF00) >> 8
        b = h & 0x0000FF

        return (r, g, b)

    augmentations = None
    if aug_config is not None:
        with open(aug_config) as file:
            config = (
                yaml.safe_load(file)
                if aug_config.suffix == ".yaml"
                else json.load(file)
            )
        augmentations = Augmentations([512, 512], config)

    dataset = LuxonisDataset(name)
    loader = LuxonisLoader(dataset, view=view.value, augmentations=augmentations)
    for image, ann in loader:
        image = image.astype(np.uint8)

        h, w, _ = image.shape

        # Ensure masks are drawn first to not occlude the other annotations
        for task, (arr, label_type) in ann.items():
            if label_type == LabelType.SEGMENTATION:
                mask_viz = np.zeros((h, w, 3)).astype(np.uint8)
                for i, mask in enumerate(arr):
                    mask_viz[mask == 1] = _task_to_rgb(f"{task}_{i}")
                image = cv2.addWeighted(image, 0.5, mask_viz, 0.5, 0)

        for task, (arr, label_type) in ann.items():
            if label_type == LabelType.BOUNDINGBOX:
                for box in arr:
                    cv2.rectangle(
                        image,
                        (int(box[1] * w), int(box[2] * h)),
                        (int(box[1] * w + box[3] * w), int(box[2] * h + box[4] * h)),
                        _task_to_rgb(task),
                        2,
                    )

            if label_type == LabelType.KEYPOINTS:
                for kp in arr:
                    kp = kp[1:].reshape(-1, 3)
                    for k in kp:
                        cv2.circle(
                            image,
                            (int(k[0] * w), int(k[1] * h)),
                            2,
                            _task_to_rgb(task),
                            2,
                        )

        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        cv2.imshow("image", image)
        if cv2.waitKey() == ord("q"):
            break


@app.command()
def parse(
    dataset_dir: Annotated[
        str, typer.Argument(..., help="Path or URL to the dataset.")
    ],
    name: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--name",
            "-n",
            autocompletion=complete_dataset_name,
            help="Name of the dataset.",
            show_default=False,
        ),
    ] = None,
    dataset_type: Annotated[
        Optional[DatasetType],
        typer.Option(
            ...,
            "--type",
            "-t",
            help="Type of the dataset. If not provided, the parser will try to recognize it automatically.",
            show_default=False,
        ),
    ] = None,
    delete_existing: Annotated[
        bool,
        typer.Option(
            ...,
            "--delete",
            "-d",
            help="If an existing dataset with the same name should "
            "be deleted before parsing.",
        ),
    ] = False,
    save_dir: Annotated[
        Optional[Path],
        typer.Option(
            ...,
            "--save-dir",
            "-s",
            help="If a remote URL is provided in 'dataset_dir', the dataset will "
            "be downloaded to this directory. Otherwise, the dataset will be "
            "downloaded to the current working directory.",
            show_default=False,
        ),
    ] = None,
):
    """Parses a directory with data and creates Luxonis dataset."""
    parser = LuxonisParser(
        dataset_dir,
        dataset_name=name,
        dataset_type=dataset_type,
        delete_existing=delete_existing,
        save_dir=save_dir,
    )
    dataset = parser.parse()

    print()
    print(Rule())
    print()
    print_info(dataset.identifier)


if __name__ == "__main__":
    app()



from abc import ABC, abstractmethod
from typing import Dict, Tuple

import numpy as np
from typing_extensions import TypeAlias

from ..utils.enums import LabelType

Labels: TypeAlias = Dict[str, Tuple[np.ndarray, LabelType]]
"""C{Labels} is a dictionary mappping task names to their L{LabelType} and annotations
as L{numpy arrays<np.ndarray>}."""


LuxonisLoaderOutput: TypeAlias = Tuple[np.ndarray, Labels]
"""C{LuxonisLoaderOutput} is a tuple of an image as a L{numpy array<np.ndarray>} and a
dictionary of task group names and their annotations as L{Annotations}."""


class BaseLoader(ABC):
    """Base abstract loader class.

    Enforces the L{LuxonisLoaderOutput} output label structure.
    """

    @abstractmethod
    def __len__(self) -> int:
        """Returns the length of the dataset.

        @rtype: int
        @return: Length of the dataset.
        """
        pass

    @abstractmethod
    def __getitem__(self, idx: int) -> LuxonisLoaderOutput:
        """Loads sample from dataset.

        @type idx: int
        @param idx: Index of the sample to load.
        @rtype: LuxonisLoaderOutput
        @return: Sample's data in L{LuxonisLoaderOutput} format.
        """
        pass



import json
import logging
import random
import warnings
from typing import Dict, List, Optional, Tuple

import cv2
import numpy as np
import polars as pl

from ..augmentations import Augmentations
from ..datasets import Annotation, LuxonisDataset, load_annotation
from ..utils.enums import BucketStorage, LabelType
from .base_loader import BaseLoader, Labels, LuxonisLoaderOutput


class LuxonisLoader(BaseLoader):
    def __init__(
        self,
        dataset: LuxonisDataset,
        view: str = "train",
        stream: bool = False,
        augmentations: Optional[Augmentations] = None,
    ) -> None:
        """A loader class used for loading data from L{LuxonisDataset}.

        @type dataset: LuxonisDataset
        @param dataset: LuxonisDataset to use
        @type view: str
        @param view: View of the dataset. Defaults to "train".
        @type stream: bool
        @param stream: Flag for data streaming. Defaults to C{False}.
        @type augmentations: Optional[luxonis_ml.loader.Augmentations]
        @param augmentations: Augmentation class that performs augmentations. Defaults
            to C{None}.
        """

        self.logger = logging.getLogger(__name__)

        self.dataset = dataset
        self.stream = stream
        self.sync_mode = (
            self.dataset.bucket_storage != BucketStorage.LOCAL and not self.stream
        )

        if self.sync_mode:
            self.logger.info("Syncing from cloud...")
            self.dataset.sync_from_cloud()
            classes_file = self.dataset.metadata_path / "classes.json"
            with open(classes_file) as file:
                synced_classes = json.load(file)
            for task in synced_classes:
                self.dataset.set_classes(classes=synced_classes[task], task=task)
            skeletons_file = self.dataset.metadata_path / "skeletons.json"
            try:
                with open(skeletons_file, "r") as file:
                    synced_skeletons = json.load(file)
                self.dataset.set_skeletons(synced_skeletons)
            except FileNotFoundError:
                self.logger.warning("Skeletons file not found at %s", skeletons_file)

        self.view = view

        self.classes, self.classes_by_task = self.dataset.get_classes(
            sync_mode=self.sync_mode
        )
        self.augmentations = augmentations
        if self.view in ["train", "val", "test"]:
            splits_path = dataset.metadata_path / "splits.json"
            if not splits_path.exists():
                raise Exception(
                    "Cannot find splits! Ensure you call dataset.make_splits()"
                )
            with open(splits_path, "r") as file:
                splits = json.load(file)
            self.instances = splits[self.view]
        else:
            raise NotImplementedError

        df = dataset._load_df_offline(sync_mode=self.sync_mode)
        if df is None:
            raise FileNotFoundError("Cannot find dataframe")
        self.df = df

        if self.dataset.bucket_storage == BucketStorage.LOCAL or not self.stream:
            file_index = self.dataset._get_file_index()
            if file_index is None:
                raise FileNotFoundError("Cannot find file index")
            self.df = self.df.join(file_index, on="uuid").drop("file_right")
        else:
            raise NotImplementedError(
                "Streaming for remote bucket storage not implemented yet"
            )

    def __len__(self) -> int:
        """Returns length of the dataset.

        @rtype: int
        @return: Length of dataset.
        """
        return len(self.instances)

    def __getitem__(self, idx: int) -> LuxonisLoaderOutput:
        """Function to load a sample consisting of an image and its annotations.

        @type idx: int
        @param idx: The (often random) integer index to retrieve a sample from the
            dataset.
        @rtype: LuxonisLoaderOutput
        @return: The loader ouput consisting of the image and a dictionary defining its
            annotations.
        """

        if self.augmentations is None:
            return self._load_image_with_annotations(idx)

        indices = [idx]
        if self.augmentations.is_batched:
            other_indices = [i for i in range(len(self)) if i != idx]
            if self.augmentations.aug_batch_size > len(self):
                warnings.warn(
                    f"Augmentations batch_size ({self.augmentations.aug_batch_size}) is larger than dataset size ({len(self)}), samples will include repetitions."
                )
                random_fun = random.choices
            else:
                random_fun = random.sample
            picked_indices = random_fun(
                other_indices, k=self.augmentations.aug_batch_size - 1
            )
            indices.extend(picked_indices)

        out_dict: Dict[str, Tuple[np.ndarray, LabelType]] = {}
        loaded_anns = [self._load_image_with_annotations(i) for i in indices]
        random_state = random.getstate()
        np_random_state = np.random.get_state()
        while loaded_anns[0][1]:
            aug_input_data = []
            label_to_task = {}
            nk = 0
            ns = 0
            for img, annotations in loaded_anns:
                label_dict: Dict[LabelType, np.ndarray] = {}
                for task in sorted(list(annotations.keys())):
                    array, label_type = annotations[task]
                    if label_type not in label_dict:
                        label_dict[label_type] = array
                        label_to_task[label_type] = task
                        annotations.pop(task)
                        if label_type == LabelType.KEYPOINTS:
                            nk = (array.shape[1] - 1) // 3
                        if label_type == LabelType.SEGMENTATION:
                            ns = array.shape[0]

                aug_input_data.append((img, label_dict))

            # NOTE: To ensure the same augmentation is applied to all samples
            # in case of multiple tasks per LabelType
            random.setstate(random_state)
            np.random.set_state(np_random_state)

            img, aug_annotations = self.augmentations(aug_input_data, nk=nk, ns=ns)
            for label_type, array in aug_annotations.items():
                out_dict[label_to_task[label_type]] = (array, label_type)

        return img, out_dict  # type: ignore

    def _load_image_with_annotations(self, idx: int) -> Tuple[np.ndarray, Labels]:
        """Loads image and its annotations based on index.

        @type idx: int
        @param idx: Index of the image
        @rtype: Tuple[L{np.ndarray}, dict]
        @return: Image as L{np.ndarray} in RGB format and a dictionary with all the
            present annotations
        """

        uuid = self.instances[idx]
        df = self.df.filter(pl.col("uuid") == uuid)
        if self.dataset.bucket_storage == BucketStorage.LOCAL:
            img_path = list(df.select("original_filepath"))[0][0]
        elif not self.stream:
            img_path = next(self.dataset.media_path.glob(f"{uuid}.*"))
        else:
            # TODO: add support for streaming remote storage
            raise NotImplementedError(
                "Streaming for remote bucket storage not implemented yet"
            )

        img = cv2.cvtColor(cv2.imread(str(img_path)), cv2.COLOR_BGR2RGB)

        height, width, _ = img.shape
        labels: Labels = {}

        for task in df["task"].unique():
            if not task:
                continue
            sub_df = df.filter(pl.col("task") == task)
            annotations: List[Annotation] = []
            class_mapping = {
                class_: i for i, class_ in enumerate(self.classes_by_task[task])
            }
            for i, (*_, type_, _, class_, instance_id, _, ann_str, _) in enumerate(
                sub_df.rows(named=False)
            ):
                instance_id = instance_id if instance_id > 0 else i
                annotation = load_annotation(
                    type_,
                    ann_str,
                    {"class": class_, "task": task, "instance_id": instance_id},
                )
                annotations.append(annotation)
            assert annotations, f"No annotations found for task {task}"
            annotations.sort(key=lambda x: x.instance_id)
            array = annotations[0].combine_to_numpy(
                annotations, class_mapping, width=width, height=height
            )
            labels[task] = (array, annotations[0]._label_type)

        return img, labels



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from albumentations import BoxType, KeypointType
from albumentations.core.bbox_utils import (
    denormalize_bbox,
    normalize_bbox,
)
from albumentations.core.transforms_interface import (
    BoxInternalType,
    ImageColorType,
    KeypointInternalType,
)

from ..batch_transform import BatchBasedTransform
from ..utils import AUGMENTATIONS


@AUGMENTATIONS.register_module()
class Mosaic4(BatchBasedTransform):
    def __init__(
        self,
        out_height: int,
        out_width: int,
        value: Optional[Union[int, float, List[int], List[float]]] = None,
        out_batch_size: int = 1,
        mask_value: Optional[Union[int, float, List[int], List[float]]] = None,
        always_apply: bool = False,
        p: float = 0.5,
    ):
        """Mosaic augmentation arranges selected four images into single image in a 2x2
        grid layout. This is done in deterministic way meaning first image in the batch
        will always be in top left. The input images should have the same number of
        channels but can have different widths and heights. The output is cropped around
        the intersection point of the four images with the size (out_with x out_height).
        If the mosaic image is smaller than with x height, the gap is filled by the
        fill_value.

        @param out_height: Output image height. The mosaic image is cropped by this height around the mosaic center.
        If the size of the mosaic image is smaller than this value the gap is filled by the `value`.
        @type out_height: int

        @param out_width: Output image width. The mosaic image is cropped by this height around the mosaic center.
        If the size of the mosaic image is smaller than this value the gap is filled by the `value`.
        @type out_width: int

        @param value: Padding value. Defaults to None.
        @type value: Optional[Union[int, float, List[int], List[float]]], optional

        @param out_batch_size: Number of output images in the batch. Defaults to 1.
        @type out_batch_size: int, optional

        @param mask_value: Padding value for masks. Defaults to None.
        @type mask_value: Optional[Union[int, float, List[int], List[float]]], optional

        @param always_apply: Whether to always apply the transform. Defaults to False.
        @type always_apply: bool, optional

        @param p: Probability of applying the transform. Defaults to 0.5.
        @type p: float, optional
        """

        super().__init__(batch_size=4, always_apply=always_apply, p=p)

        if out_height <= 0:
            raise ValueError(f"out_height should be larger than 0, got {out_height}")
        if out_width <= 0:
            raise ValueError(f"out_width should be larger than 0, got {out_width}")
        if out_batch_size <= 0:
            raise ValueError(
                f"out_batch_size should be larger than 0, got {out_batch_size}"
            )

        self.n_tiles = self.batch_size  # 4: 2x2
        self.out_height = out_height
        self.out_width = out_width
        self.value = value
        self.mask_value = mask_value
        self.out_batch_size = out_batch_size

    def get_transform_init_args_names(self) -> Tuple[str, ...]:
        """Gets the default arguments for the mixup augmentation.

        @return: The string keywords of the arguments.
        @rtype: Tuple[str, ...]
        """
        return (
            "out_height",
            "out_width",
            "replace",
            "value",
            "out_batch_size",
            "mask_value",
        )

    @property
    def targets_as_params(self):
        """List of augmentation targets.

        @return: Output list of augmentation targets.
        @rtype: List[str]
        """
        return ["image_batch"]

    def apply_to_image_batch(
        self, image_batch: List[np.ndarray], indices: List[int], **params
    ) -> List[np.ndarray]:
        """Applies the transformation to a batch of images.

        @param image_batch: Batch of input images to which the transformation is
            applied.
        @type image_batch: List[np.ndarray]
        @param indices: Indices of images in the batch.
        @type indices: List[Tuple[int, int]]
        @param params: Additional parameters for the transformation.
        @type params: Any
        @return: List of transformed images.
        @rtype: List[np.ndarray]
        """
        output_batch = []
        for i_batch in range(self.out_batch_size):
            idx_chunk = indices[self.n_tiles * i_batch : self.n_tiles * (i_batch + 1)]
            image_chunk = [image_batch[i] for i in idx_chunk]
            mosaiced = mosaic4(image_chunk, self.out_height, self.out_width, self.value)
            output_batch.append(mosaiced)
        return output_batch

    def apply_to_mask_batch(
        self, mask_batch: List[np.ndarray], indices: List[int], **params
    ) -> List[np.ndarray]:
        """Applies the transformation to a batch of masks.

        @param mask_batch: Batch of input masks to which the transformation is applied.
        @type mask_batch: List[np.ndarray]
        @param indices: Indices of images in the batch.
        @type indices: List[Tuple[int, int]]
        @param params: Additional parameters for the transformation.
        @type params: Any
        @return: List of transformed masks.
        @rtype: List[np.ndarray]
        """
        output_batch = []
        for i_batch in range(self.out_batch_size):
            idx_chunk = indices[self.n_tiles * i_batch : self.n_tiles * (i_batch + 1)]
            mask_chunk = [mask_batch[i] for i in idx_chunk]
            mosaiced = mosaic4(
                mask_chunk, self.out_height, self.out_width, self.mask_value
            )
            output_batch.append(mosaiced)
        return output_batch

    def apply_to_bboxes_batch(
        self,
        bboxes_batch: List[BoxType],
        indices: List[int],
        image_shapes: List[Tuple[int, int]],
        **params,
    ) -> List[BoxType]:
        """Applies the transformation to a batch of bboxes.

        @param bboxes_batch: Batch of input bboxes to which the transformation is
            applied.
        @type bboxes_batch: List[BboxType]
        @param indices: Indices of images in the batch.
        @type indices: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type image_shapes: List[Tuple[int, int]]
        @param params: Additional parameters for the transformation.
        @type params: Any
        @return: List of transformed bboxes.
        @rtype: List[BoxType]
        """
        output_batch = []
        for i_batch in range(self.out_batch_size):
            idx_chunk = indices[self.n_tiles * i_batch : self.n_tiles * (i_batch + 1)]
            bboxes_chunk = [bboxes_batch[i] for i in idx_chunk]
            shape_chunk = [image_shapes[i] for i in idx_chunk]
            new_bboxes = []
            for i in range(self.n_tiles):
                bboxes = bboxes_chunk[i]
                rows, cols = shape_chunk[i]
                for bbox in bboxes:
                    new_bbox = bbox_mosaic4(
                        bbox[:4], rows, cols, i, self.out_height, self.out_width
                    )
                    new_bboxes.append(tuple(new_bbox) + tuple(bbox[4:]))
            output_batch.append(new_bboxes)
        return output_batch

    def apply_to_keypoints_batch(
        self,
        keyboints_batch: List[KeypointType],
        indices: List[int],
        image_shapes: List[Tuple[int, int]],
        **params,
    ) -> List[KeypointType]:
        """Applies the transformation to a batch of keypoints.

        @param keypoints_batch: Batch of input keypoints to which the transformation is
            applied.
        @type keypoints_batch: List[KeypointType]
        @param indices: Indices of images in the batch.
        @type indices: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type image_shapes: List[Tuple[int, int]]
        @param params: Additional parameters for the transformation.
        @type params: Any
        @return: List of transformed keypoints.
        @rtype: List[KeypointType]
        """
        output_batch = []
        for i_batch in range(self.out_batch_size):
            idx_chunk = indices[self.n_tiles * i_batch : self.n_tiles * (i_batch + 1)]
            keypoints_chunk = [keyboints_batch[i] for i in idx_chunk]
            shape_chunk = [image_shapes[i] for i in idx_chunk]
            new_keypoints = []
            for i in range(self.n_tiles):
                keypoints = keypoints_chunk[i]
                rows, cols = shape_chunk[i]
                for keypoint in keypoints:
                    new_keypoint = keypoint_mosaic4(
                        keypoint[:4], rows, cols, i, self.out_height, self.out_width
                    )
                    new_keypoints.append(new_keypoint + tuple(keypoint[4:]))
            output_batch.append(new_keypoints)
        return output_batch

    def get_params_dependent_on_targets(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Get parameters dependent on the targets.

        @param params: Dictionary containing parameters.
        @type params: Dict[str, Any]
        @return: Dictionary containing parameters dependent on the targets.
        @rtype: Dict[str, Any]
        """
        image_batch = params["image_batch"]
        n = len(image_batch)
        if self.n_tiles * self.out_batch_size > n:
            raise ValueError(
                f"The batch size (= {n}) should be larger than "
                + f"{self.n_tiles} x out_batch_size (= {self.n_tiles * self.out_batch_size})"
            )
        indices = np.random.choice(
            range(n), size=self.n_tiles * self.out_batch_size, replace=False
        ).tolist()
        image_shapes = [tuple(image.shape[:2]) for image in image_batch]
        return {
            "indices": indices,
            "image_shapes": image_shapes,
        }


def mosaic4(
    image_batch: List[np.ndarray],
    height: int,
    width: int,
    value: Optional[ImageColorType] = None,
) -> np.ndarray:
    """Arrange the images in a 2x2 grid layout.
    The input images should have the same number of channels but can have different widths and heights.
    The output is cropped around the intersection point of the four images with the size (with x height).
    If the mosaic image is smaller than with x height, the gap is filled by the fill_value.
    This implementation is based on YOLOv5 with some modification:
    https://github.com/ultralytics/yolov5/blob/932dc78496ca532a41780335468589ad7f0147f7/utils/datasets.py#L648

    @param image_batch: Image list. The length should be four. Each image can has different size.
    @type image_batch: List[np.ndarray]
    @param height: Height of output mosaic image
    @type height: int
    @param width: Width of output mosaic image
    @type width: int
    @param value: Padding value
    @type value: Optional[ImageColorType]
    @return: Final output image
    @rtype: np.ndarray
    """
    N_TILES = 4
    if len(image_batch) != N_TILES:
        raise ValueError(f"Length of image_batch should be 4. Got {len(image_batch)}")

    for i in range(N_TILES - 1):
        if image_batch[0].shape[2:] != image_batch[i + 1].shape[2:]:
            raise ValueError(
                "All images should have the same number of channels."
                + f" Got the shapes {image_batch[0].shape} and {image_batch[i + 1].shape}"
            )

        if image_batch[0].dtype != image_batch[i + 1].dtype:
            raise ValueError(
                "All images should have the same dtype."
                + f" Got the dtypes {image_batch[0].dtype} and {image_batch[i + 1].dtype}"
            )

    if len(image_batch[0].shape) == 2:
        out_shape = [height, width]
    else:
        out_shape = [height, width, image_batch[0].shape[2]]

    dtype = image_batch[0].dtype
    img4 = np.zeros(out_shape, dtype=dtype)  # base image with 4 tiles

    value = 0 if value is None else value
    if isinstance(value, (tuple, list, np.ndarray)):
        if out_shape[2] != len(value):
            ValueError(
                "value parameter should has the same lengh as the output channel."
                + f" value: ({value}), output shape: {out_shape}"
            )
        for i in range(len(value)):
            img4[:, :, i] = value[i]
    else:
        img4[:] = value

    center_x = width // 2
    center_y = height // 2
    for i, img in enumerate(image_batch):
        (h, w) = img.shape[:2]

        # place img in img4
        # this based on the yolo5's implementation
        #
        if i == 0:  # top left
            x1a, y1a, x2a, y2a = (
                max(center_x - w, 0),
                max(center_y - h, 0),
                center_x,
                center_y,
            )  # xmin, ymin, xmax, ymax (large image)
            x1b, y1b, x2b, y2b = (
                w - (x2a - x1a),
                h - (y2a - y1a),
                w,
                h,
            )  # xmin, ymin, xmax, ymax (small image)
        elif i == 1:  # top right
            x1a, y1a, x2a, y2a = (
                center_x,
                max(center_y - h, 0),
                min(center_x + w, width),
                center_y,
            )
            x1b, y1b, x2b, y2b = 0, h - (y2a - y1a), min(w, x2a - x1a), h
        elif i == 2:  # bottom left
            x1a, y1a, x2a, y2a = (
                max(center_x - w, 0),
                center_y,
                center_x,
                min(height, center_y + h),
            )
            x1b, y1b, x2b, y2b = w - (x2a - x1a), 0, w, min(y2a - y1a, h)
        elif i == 3:  # bottom right
            x1a, y1a, x2a, y2a = (
                center_x,
                center_y,
                min(center_x + w, width),
                min(height, center_y + h),
            )
            x1b, y1b, x2b, y2b = 0, 0, min(w, x2a - x1a), min(y2a - y1a, h)

        img4[y1a:y2a, x1a:x2a] = img[y1b:y2b, x1b:x2b]  # img4[ymin:ymax, xmin:xmax]

    return img4


def bbox_mosaic4(
    bbox: BoxInternalType,
    rows: int,
    cols: int,
    position_index: int,
    height: int,
    width: int,
) -> BoxInternalType:
    """Put the given bbox in one of the cells of the 2x2 grid.

    @param bbox: A bounding box `(x_min, y_min, x_max, y_max)`.
    @type bbox: BoxInternalType
    @param rows: Height of input image that corresponds to one of the mosaic cells
    @type rows: int
    @param cols: Width of input image that corresponds to one of the mosaic cells
    @type cols: int
    @param position_index: Index of the mosaic cell. 0: top left, 1: top right, 2:
        bottom left, 3: bottom right
    @type position_index: int
    @param height: Height of output mosaic image
    @type height: int
    @param width: Width of output mosaic image
    @type width: int
    @return: Transformed bbox
    @rtype: BoxInternalType
    """
    bbox = denormalize_bbox(bbox, rows, cols)
    center_x = width // 2
    center_y = height // 2
    if position_index == 0:  # top left
        shift_x = center_x - cols
        shift_y = center_y - rows
    elif position_index == 1:  # top right
        shift_x = center_x
        shift_y = center_y - rows
    elif position_index == 2:  # bottom left
        shift_x = center_x - cols
        shift_y = center_y
    elif position_index == 3:  # bottom right
        shift_x = center_x
        shift_y = center_y
    bbox = (
        bbox[0] + shift_x,
        bbox[1] + shift_y,
        bbox[2] + shift_x,
        bbox[3] + shift_y,
    )
    bbox = normalize_bbox(bbox, height, width)
    return bbox


def keypoint_mosaic4(
    keypoint: KeypointInternalType,
    rows: int,
    cols: int,
    position_index: int,
    height: int,
    width: int,
) -> KeypointInternalType:
    """Put the given keypoint in one of the cells of the 2x2 grid.

    @param keypoint: A keypoint `(x, y, angle, scale)`.
    @type bbox: KeypointInternalType
    @param rows: Height of input image that corresponds to one of the mosaic cells
    @type rows: int
    @param cols: Width of input image that corresponds to one of the mosaic cells
    @type cols: int
    @param position_index: Index of the mosaic cell. 0: top left, 1: top right, 2:
        bottom left, 3: bottom right
    @type position_index: int
    @param height: Height of output mosaic image
    @type height: int
    @param width: Width of output mosaic image
    @type width: int
    @return: Transformed keypoint
    @rtype: KeypointInternalType
    """
    x, y, angle, scale = keypoint

    center_x = width // 2
    center_y = height // 2
    if position_index == 0:  # top left
        shift_x = center_x - cols
        shift_y = center_y - rows
    elif position_index == 1:  # top right
        shift_x = center_x
        shift_y = center_y - rows
    elif position_index == 2:  # bottom left
        shift_x = center_x - cols
        shift_y = center_y
    elif position_index == 3:  # bottom right
        shift_x = center_x
        shift_y = center_y
    return x + shift_x, y + shift_y, angle, scale



from enum import Enum


class DataType(Enum):
    """Represents all existing data types used in i/o streams of the model."""

    INT8 = "int8"
    INT32 = "int32"
    UINT8 = "uint8"
    FLOAT32 = "float32"
    FLOAT16 = "float16"



import logging
import sys
from contextlib import contextmanager

logger = logging.getLogger(__name__)


@contextmanager
def guard_missing_extra(name: str):
    try:
        yield
    except ImportError:
        logger.exception(
            f"Error importing `luxonis-ml.{name}`. This can mean that some of the dependencies of `luxonis-ml[{name}]` are not installed. "
            f"Ensure you installed the package with the `[{name}]` or `[all]` extra specified. "
            f"Use `pip install luxonis-ml[{name}]` to install dependencies for the `{name}` submodule."
        )
        sys.exit(1)


__all__ = ["guard_missing_extra"]



import os
import xml.etree.ElementTree as ET
from typing import Tuple

import numpy as np

from .. import LuxonisDataset


def generate_random_color() -> Tuple[int, int, int]:
    """Generates a random RGB color for labelstudio visualization.

    @rtype: Tuple[int, int, int]
    @return: A randomly generated RGB value.
    """

    return (np.random.randint(0, 255) for _ in range(3))


def get_xml_config(dataset: LuxonisDataset) -> str:
    """Generates the labelstudio XML for visualizing an annotation project for a
    LuxonisDataset.

    @type dataset: C{LuxonisDataset}
    @param dataset: The LuxonisDataset used to generate the XML
    @rtype: str
    @return: The XML file contents as a string.
    """

    root = ET.Element("View")

    image = ET.SubElement(root, "Image")
    image.set("name", "image")
    image.set("value", "$image")
    image.set("zoom", "true")

    _, classes = dataset.get_classes()
    if "class" in classes and len(classes["class"]):
        choices = ET.SubElement(root, "Choices")
        choices.set("name", "choice")
        choices.set("toName", "image")
        for label in classes["class"]:
            choice = ET.SubElement(choices, "Choice")
            choice.set("value", label)

    if "boxes" in classes and len(classes["boxes"]):
        inst = ET.SubElement(root, "RectangleLabels")
        inst.set("name", "label")
        inst.set("toName", "image")
        for label in classes["boxes"]:
            r, g, b = generate_random_color()
            lbl = ET.SubElement(inst, "Label")
            lbl.set("value", label)
            lbl.set("background", f"rgba({r},{b},{g},0.7)")

    if "segmentation" in classes and len(classes["segmentation"]):
        inst = ET.SubElement(root, "BrushLabels")
        inst.set("name", "tag")
        inst.set("toName", "image")
        for label in classes["segmentation"]:
            r, g, b = generate_random_color()
            lbl = ET.SubElement(inst, "Label")
            lbl.set("value", label)
            lbl.set("background", f"rgba({r},{b},{g},0.7)")

    if "keypoints" in classes and len(classes["keypoints"]):
        inst = ET.SubElement(root, "KeyPointLabels")
        inst.set("name", "kp-1")
        inst.set("toName", "image")
        for label in classes["keypoints"]:
            r, g, b = generate_random_color()
            lbl = ET.SubElement(inst, "Label")
            lbl.set("value", label)
            lbl.set("background", f"rgba({r},{b},{g},0.7)")

    tree = ET.ElementTree(root)
    tree.write("tmp.xml", encoding="utf-8", xml_declaration=False)

    with open("tmp.xml") as file:
        label_config_str = file.read()

    os.remove("tmp.xml")

    return label_config_str



import json
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Literal, Optional, Tuple, TypedDict, Union

import numpy as np
import pycocotools.mask as mask_util
from PIL import Image, ImageDraw
from pydantic import Field, model_validator
from pydantic.types import FilePath, PositiveInt
from typing_extensions import Annotated, TypeAlias

from luxonis_ml.utils import BaseModelExtraForbid, Registry

from ..utils.enums import LabelType

DATASETS_REGISTRY = Registry(name="datasets")


KeypointVisibility: TypeAlias = Literal[0, 1, 2]
NormalizedFloat: TypeAlias = Annotated[float, Field(ge=0, le=1)]
"""C{NormalizedFloat} is a float that is restricted to the range [0, 1]."""

ParquetDict = TypedDict(
    "ParquetDict",
    {
        "file": str,
        "type": str,
        "created_at": datetime,
        "class": str,
        "instance_id": int,
        "task": str,
        "annotation": str,
    },
)


def load_annotation(name: str, js: str, data: Dict[str, Any]) -> "Annotation":
    return {
        "ClassificationAnnotation": ClassificationAnnotation,
        "BBoxAnnotation": BBoxAnnotation,
        "KeypointAnnotation": KeypointAnnotation,
        "RLESegmentationAnnotation": RLESegmentationAnnotation,
        "PolylineSegmentationAnnotation": PolylineSegmentationAnnotation,
        "ArrayAnnotation": ArrayAnnotation,
        "LabelAnnotation": LabelAnnotation,
    }[name](**json.loads(js), **data)


class Annotation(ABC, BaseModelExtraForbid):
    """Base class for an annotation.

    @type task: str
    @ivar task: The task name. By default it is the string representation of the
        L{LabelType}.
    @type class_: str
    @ivar class_: The class name for the annotation.
    @type instance_id: int
    @ivar instance_id: The instance id of the annotation. This determines the order in
        which individual instances are loaded in L{LuxonisLoader}.
    @type _label_type: ClassVar[L{LabelType}]
    @ivar _label_type: The label type of the annotation.
    """

    _label_type: ClassVar[LabelType]

    task: str
    class_: str = Field("", alias="class")
    instance_id: int = -1

    @model_validator(mode="before")
    @classmethod
    def validate_task(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if "task" not in values:
            values = {**values, "task": cls._label_type.value}
        return values

    def get_value(self) -> Dict[str, Any]:
        """Converts the annotation to a dictionary that can be saved to a parquet
        file."""
        return self.dict(exclude={"class_", "class_id", "instance_id", "task", "type_"})

    @staticmethod
    @abstractmethod
    def combine_to_numpy(
        annotations: List["Annotation"],
        class_mapping: Dict[str, int],
        height: int,
        width: int,
    ) -> np.ndarray:
        """Combines multiple instance annotations into a single numpy array."""
        pass


class ClassificationAnnotation(Annotation):
    _label_type = LabelType.CLASSIFICATION
    type_: Literal["classification"] = Field("classification", alias="type")

    @staticmethod
    def combine_to_numpy(
        annotations: List["ClassificationAnnotation"],
        class_mapping: Dict[str, int],
        **_,
    ) -> np.ndarray:
        classify_vector = np.zeros(len(class_mapping))
        for ann in annotations:
            class_ = class_mapping.get(ann.class_, 0)
            classify_vector[class_] = 1
        return classify_vector


class BBoxAnnotation(Annotation):
    """Bounding box annotation.

    Values are normalized based on the image size.

    @type x: float
    @ivar x: The center x-coordinate of the bounding box. Normalized to [0, 1].
    @type y: float
    @ivar y: The center y-coordinate of the bounding box. Normalized to [0, 1].
    @type w: float
    @ivar w: The width of the bounding box. Normalized to [0, 1].
    @type h: float
    @ivar h: The height of the bounding box. Normalized to [0, 1].
    """

    type_: Literal["boundingbox"] = Field("boundingbox", alias="type")

    x: NormalizedFloat
    y: NormalizedFloat
    w: NormalizedFloat
    h: NormalizedFloat

    _label_type = LabelType.BOUNDINGBOX

    def to_numpy(self, class_mapping: Dict[str, int]) -> np.ndarray:
        class_ = class_mapping.get(self.class_, 0)
        return np.array([class_, self.x, self.y, self.w, self.h])

    @staticmethod
    def combine_to_numpy(
        annotations: List["BBoxAnnotation"], class_mapping: Dict[str, int], **_
    ) -> np.ndarray:
        boxes = np.zeros((len(annotations), 5))
        for i, ann in enumerate(annotations):
            boxes[i] = ann.to_numpy(class_mapping)
        return boxes


class KeypointAnnotation(Annotation):
    """Keypoint annotation.

    Values are normalized to [0, 1] based on the image size.

    @type keypoints: List[Tuple[float, float, L{KeypointVisibility}]]
    @ivar keypoints: List of keypoints. Each keypoint is a tuple of (x, y, visibility).
        x and y are normalized to [0, 1]. visibility is one of {0, 1, 2} where:
            - 0: Not visible / not labeled
            - 1: Occluded
            - 2: Visible
    """

    type_: Literal["keypoints"] = Field("keypoints", alias="type")

    keypoints: List[Tuple[NormalizedFloat, NormalizedFloat, KeypointVisibility]]

    _label_type = LabelType.KEYPOINTS

    def to_numpy(self, class_mapping: Dict[str, int]) -> np.ndarray:
        class_ = class_mapping.get(self.class_, 0)
        kps = np.array(self.keypoints).reshape((-1, 3)).astype(np.float32)
        return np.concatenate([[class_], kps.flatten()])

    @staticmethod
    def combine_to_numpy(
        annotations: List["KeypointAnnotation"], class_mapping: Dict[str, int], **_
    ) -> np.ndarray:
        keypoints = np.zeros((len(annotations), len(annotations[0].keypoints) * 3 + 1))
        for i, ann in enumerate(annotations):
            keypoints[i] = ann.to_numpy(class_mapping)
        return keypoints


class SegmentationAnnotation(Annotation):
    """Base class for segmentation annotations."""

    _label_type = LabelType.SEGMENTATION


class RLESegmentationAnnotation(SegmentationAnnotation):
    """U{Run-length encoded<https://en.wikipedia.org/wiki/Run-length_encoding>}
        segmentation mask.

    @type height: int
    @ivar height: The height of the segmentation mask.

    @type width: int
    @ivar width: The width of the segmentation mask.

    @type counts: Union[List[int], bytes]
    @ivar counts: The run-length encoded mask.
        This can be a list of integers or a byte string.
    """

    type_: Literal["rle"] = Field("rle", alias="type")

    height: PositiveInt
    width: PositiveInt
    counts: Union[List[PositiveInt], bytes]

    def get_value(self) -> Dict[str, Any]:
        if isinstance(self.counts, bytes):
            return {
                "height": self.height,
                "width": self.width,
                "counts": self.counts.decode("utf-8"),
            }

        rle: Any = mask_util.frPyObjects(
            {
                "counts": self.counts,
                "size": [self.height, self.width],
            },
            self.height,
            self.width,
        )

        return {
            "height": rle["size"][0],
            "width": rle["size"][1],
            "counts": rle["counts"].decode("utf-8"),
        }

    def to_numpy(self, _: Dict[str, int], width: int, height: int) -> np.ndarray:
        assert isinstance(self.counts, bytes)
        return mask_util.decode({"counts": self.counts, "size": [height, width]})

    @staticmethod
    def combine_to_numpy(
        annotations: List["RLESegmentationAnnotation"],
        class_mapping: Dict[str, int],
        height: int,
        width: int,
    ) -> np.ndarray:
        seg = np.zeros((len(class_mapping), height, width))
        for ann in annotations:
            class_ = class_mapping.get(ann.class_, 0)
            mask = ann.to_numpy(class_mapping, width, height)
            seg[class_, ...] += mask

        seg = np.clip(seg, 0, 1)
        return seg


class PolylineSegmentationAnnotation(SegmentationAnnotation):
    """Polyline segmentation mask.

    @type points: List[Tuple[float, float]]
    @ivar points: List of points that define the polyline. Each point is a tuple of (x,
        y). x and y are normalized to [0, 1] based on the image size.
    """

    type_: Literal["polyline"] = Field("polyline", alias="type")

    points: List[Tuple[NormalizedFloat, NormalizedFloat]] = Field(min_length=3)

    def to_numpy(self, _: Dict[str, int], width: int, height: int) -> np.ndarray:
        polyline = [(round(x * width), round(y * height)) for x, y in self.points]
        mask = Image.new("L", (width, height), 0)
        draw = ImageDraw.Draw(mask)
        draw.polygon(polyline, fill=1, outline=1)
        return np.array(mask)

    @staticmethod
    def combine_to_numpy(
        annotations: List["PolylineSegmentationAnnotation"],
        class_mapping: Dict[str, int],
        height: int,
        width: int,
    ) -> np.ndarray:
        seg = np.zeros((len(class_mapping), height, width))
        for ann in annotations:
            class_ = class_mapping.get(ann.class_, 0)
            seg[class_, ...] += ann.to_numpy(class_mapping, width, height)

        seg = np.clip(seg, 0, 1)
        return seg


class ArrayAnnotation(Annotation):
    """A custom unspecified annotation that is an arbitrary numpy array.

    All instances of this annotation must have the same shape.

    @type path: L{FilePath}
    @ivar path: The path to the numpy array saved as a C{.npy} file.
    """

    type_: Literal["array"] = Field("array", alias="type")

    path: FilePath

    _label_type = LabelType.ARRAY

    @staticmethod
    def combine_to_numpy(
        annotations: List["ArrayAnnotation"], class_mapping: Dict[str, int], **_
    ) -> np.ndarray:
        out_arr = np.zeros(
            (len(annotations), len(class_mapping), *np.load(annotations[0].path).shape)
        )
        for i, ann in enumerate(annotations):
            class_ = class_mapping.get(ann.class_, 0)
            out_arr[i, class_] = np.load(ann.path)
        return out_arr


class LabelAnnotation(Annotation):
    """A custom unspecified annotation with a single primitive value.

    @type value: Union[bool, int, float, str]
    @ivar value: The value of the annotation.
    """

    type_: Literal["label"] = Field("label", alias="type")

    value: Union[bool, int, float, str]

    _label_type = LabelType.LABEL

    @staticmethod
    def combine_to_numpy(
        annotations: List["LabelAnnotation"], class_mapping: Dict[str, int], **_
    ) -> np.ndarray:
        out_arr = np.zeros((len(annotations), len(class_mapping))).astype(
            type(annotations[0].value)
        )
        for i, ann in enumerate(annotations):
            class_ = class_mapping.get(ann.class_, 0)
            out_arr[i, class_] = ann.value
        return out_arr


class DatasetRecord(BaseModelExtraForbid):
    """A record of an image and its annotation.

    @type file: L{FilePath}
    @ivar file: A path to the image.
    @type annotation: Optional[Annotation]
    @ivar annotation: The annotation for the image.
    """

    file: FilePath
    annotation: Optional[
        Union[
            ClassificationAnnotation,
            BBoxAnnotation,
            KeypointAnnotation,
            RLESegmentationAnnotation,
            PolylineSegmentationAnnotation,
            ArrayAnnotation,
            LabelAnnotation,
        ]
    ] = Field(None, discriminator="type_")

    def to_parquet_dict(self) -> ParquetDict:
        """Converts an annotation to a dictionary for writing to a parquet file.

        @rtype: L{ParquetDict}
        @return: A dictionary of annotation data.
        """

        value = self.annotation.get_value() if self.annotation is not None else {}
        json_value = json.dumps(value)
        return {
            "file": self.file.name,
            "type": self.annotation.__class__.__name__,
            "created_at": datetime.utcnow(),
            "class": self.annotation.class_ or ""
            if self.annotation is not None
            else "",
            "instance_id": self.annotation.instance_id or -1
            if self.annotation is not None
            else -1,
            "task": self.annotation.task if self.annotation is not None else "",
            "annotation": json_value,
        }



from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple, Union

from typing_extensions import TypeAlias

from luxonis_ml.utils import AutoRegisterMeta, Registry
from luxonis_ml.utils.filesystem import PathType

from .annotation import DatasetRecord
from .source import LuxonisSource

DATASETS_REGISTRY = Registry(name="datasets")


DatasetIterator: TypeAlias = Iterator[Union[dict, DatasetRecord]]


class BaseDataset(
    ABC, metaclass=AutoRegisterMeta, registry=DATASETS_REGISTRY, register=False
):
    """Base abstract dataset class for managing datasets in the Luxonis MLOps
    ecosystem."""

    @property
    @abstractmethod
    def identifier(self) -> str:
        """The unique identifier for the dataset."""
        pass

    @abstractmethod
    def get_tasks(self) -> List[str]:
        """Returns the list of tasks in the dataset.

        @rtype: List[str]
        @return: List of task names.
        """
        pass

    @abstractmethod
    def update_source(self, source: LuxonisSource) -> None:
        """Updates underlying source of the dataset with a new LuxonisSource.

        @type source: L{LuxonisSource}
        @param source: The new L{LuxonisSource} to replace the old one.
        """
        pass

    @abstractmethod
    def set_classes(self, classes: List[str], task: Optional[str] = None) -> None:
        """Sets the names of classes for the dataset. This can be across all CV tasks or
        certain tasks.

        @type classes: List[str]
        @param classes: List of class names to set.
        @type task: Optional[str]
        @param task: Optionally specify the task where these classes apply.
        """
        pass

    @abstractmethod
    def get_classes(self) -> Tuple[List[str], Dict[str, List[str]]]:
        """Gets overall classes in the dataset and classes according to computer vision
        task.

        @type sync_mode: bool
        @param sync_mode: If C{True}, reads classes from remote storage. If C{False},
            classes are read locally.
        @rtype: Tuple[List[str], Dict]
        @return: A combined list of classes for all tasks and a dictionary mapping tasks
            to the classes used in each task.
        """
        pass

    @abstractmethod
    def set_skeletons(
        self, skeletons: Dict[str, Dict[str, Any]], task: Optional[str] = None
    ) -> None:
        """Sets the semantic structure of keypoint skeletons for the classes that use
        keypoints.

        @type skeletons: Dict[str, Dict]
        @param skeletons: A dict mapping class name to keypoint "labels" and "edges"
            between keypoints.
            The inclusion of "edges" is optional.

            Example::

                {
                    "person": {
                        "labels": ["right hand", "right shoulder", ...]
                        "edges" [[0, 1], [4, 5], ...]
                    }
                }
        """
        pass

    @abstractmethod
    def get_skeletons(self) -> Dict[str, Dict]:
        """Returns the dictionary defining the semantic skeleton for each class using
        keypoints.

        @rtype: Dict[str, Dict]
        @return: A dictionary mapping classes to their skeleton definitions.
        """
        pass

    @abstractmethod
    def add(self, generator: DatasetIterator, batch_size: int = 1_000_000) -> None:
        """Write annotations to parquet files.

        @type generator: L{DatasetGenerator}
        @param generator: A Python iterator that yields either instances of
            L{DatasetRecord} or a dictionary that can be converted to L{DatasetRecord}.
        @type batch_size: int
        @param batch_size: The number of annotations generated before processing. This
            can be set to a lower value to reduce memory usage.
        """
        pass

    # TODO: Support arbitrary named splits
    @abstractmethod
    def make_splits(
        self,
        ratios: Tuple[float, float, float] = (0.8, 0.1, 0.1),
        definitions: Optional[Dict[str, Sequence[PathType]]] = None,
    ) -> None:
        """Saves a splits json file that specified the train/val/test split. For use in
        I{OFFLINE} mode only.

        @type ratios: Tuple[float, float, float]
            Defaults to (0.8, 0.1, 0.1).

        @type definitions: Optional[Dict]
        @param definitions [Optional[Dict]]: Dictionary specifying split keys to lists
            of filepath values. Note that this assumes unique filenames.
            Example::

                {
                    "train": ["/path/to/cat.jpg", "/path/to/dog.jpg"],
                    "val": [...],
                    "test": [...]
                }

            Only overrides splits that are present in the dictionary.
        """
        pass

    @abstractmethod
    def delete_dataset(self) -> None:
        """Deletes all local files belonging to the dataset."""
        pass

    @staticmethod
    @abstractmethod
    def exists(dataset_name: str) -> bool:
        """Checks whether a dataset exists.

        @warning: For offline mode only.
        @type dataset_name: str
        @param dataset_name: Name of the dataset
        @rtype: bool
        @return: Whether the dataset exists
        """
        pass



from pydantic import Field

from luxonis_ml.utils import BaseModelExtraForbid

from ..enums import DataType


class Output(BaseModelExtraForbid):
    """Represents output stream of a model.

    @type name: str
    @ivar name: Name of the output layer.
    @type dtype: DataType
    @ivar dtype: Data type of the output data (e.g., 'float32').
    """

    name: str = Field(description="Name of the output layer.")
    dtype: DataType = Field(
        description="Data type of the output data (e.g., 'float32')."
    )



import tarfile
from pathlib import Path

from luxonis_ml.utils.filesystem import PathType


def is_nn_archive(path: PathType) -> bool:
    """Check if the given path is a valid NN archive file.

    @type path: PathType
    @param path: Path to the file to check.
    @rtype: bool
    @return: True if the file is a valid NN archive file, False otherwise.
    """
    path = Path(path)

    if not path.is_file():
        return False

    if not tarfile.is_tarfile(path):
        return False

    with tarfile.open(path, "r") as tar:
        if "config.json" not in tar.getnames():
            return False

    return True



from .data_type import DataType
from .decoding import ObjectDetectionSubtypeSSD, ObjectDetectionSubtypeYOLO
from .input_type import InputType

__all__ = [
    "DataType",
    "InputType",
    "ObjectDetectionSubtypeSSD",
    "ObjectDetectionSubtypeYOLO",
]



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class COCOParser(BaseParser):
    """Parses directory with COCO annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── data/
        │   │   ├── img1.jpg
        │   │   ├── img2.jpg
        │   │   └── ...
        │   └── labels.json
        ├── validation/
        │   ├── data/
        │   └── labels.json
        └── test/
            ├── data/
            └── labels.json

    This is default format returned when using FiftyOne package.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None
        json_path = next(split_path.glob("*.json"), None)
        if not json_path:
            return None
        data_path = split_path / json_path.stem
        if not data_path.exists():
            dirs = [d for d in split_path.iterdir() if d.is_dir()]
            if len(dirs) != 1:
                return None
            data_path = dirs[0]
        return {"image_dir": data_path, "annotation_path": json_path}

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "validation", "test"]:
            split_path = dataset_dir / split
            if COCOParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self,
        dataset_dir: Path,
        use_keypoint_ann: bool = False,
        keypoint_ann_paths: Optional[Dict[str, str]] = None,
        split_val_to_test: bool = True,
    ) -> Tuple[List[str], List[str], List[str]]:
        if use_keypoint_ann and not keypoint_ann_paths:
            keypoint_ann_paths = {
                "train": "raw/person_keypoints_train2017.json",
                "val": "raw/person_keypoints_val2017.json",
                # NOTE: this file is not present by default
                "test": "raw/person_keypoints_test2017.json",
            }

        train_ann_path = (
            dataset_dir / keypoint_ann_paths["train"]
            if keypoint_ann_paths and use_keypoint_ann
            else dataset_dir / "train" / "labels.json"
        )
        added_train_imgs = self._parse_split(
            image_dir=dataset_dir / "train" / "data",
            annotation_path=train_ann_path,
        )

        val_ann_path = (
            dataset_dir / keypoint_ann_paths["val"]
            if keypoint_ann_paths and use_keypoint_ann
            else dataset_dir / "validation" / "labels.json"
        )
        _added_val_imgs = self._parse_split(
            image_dir=dataset_dir / "validation" / "data",
            annotation_path=val_ann_path,
        )

        if split_val_to_test:
            split_point = round(len(_added_val_imgs) * 0.5)
            added_val_imgs = _added_val_imgs[:split_point]
            added_test_imgs = _added_val_imgs[split_point:]
        else:
            added_val_imgs = _added_val_imgs
            added_test_imgs = []
            # NOTE: test split annotations are not included by default
            test_ann_path = (
                dataset_dir / keypoint_ann_paths["test"]
                if keypoint_ann_paths and use_keypoint_ann
                else dataset_dir / "test" / "labels.json"
            )
            if test_ann_path.exists():
                added_test_imgs = self._parse_split(
                    image_dir=dataset_dir / "test" / "data",
                    annotation_path=test_ann_path,
                )

        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(self, image_dir: Path, annotation_path: Path) -> ParserOutput:
        """Parses annotations from COCO format to LDF. Annotations include
        classification, segmentation, object detection and keypoints if present.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type annotation_path: Path
        @param annotation_path: Path to annotation json file
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton dictionary for
            keypoints and list of added images.
        """
        with open(annotation_path) as f:
            annotation_data = json.load(f)

        coco_images = annotation_data["images"]
        coco_annotations = annotation_data["annotations"]
        coco_categories = annotation_data["categories"]
        categories = {cat["id"]: cat["name"] for cat in coco_categories}

        class_names = list(categories.values())
        skeletons = {}
        for cat in coco_categories:
            if "keypoints" in cat.keys() and "skeleton" in cat.keys():
                skeletons[categories[cat["id"]]] = {
                    "labels": cat["keypoints"],
                    "edges": (np.array(cat["skeleton"]) - 1).tolist(),
                }

        def generator() -> DatasetIterator:
            img_dict = {img["id"]: img for img in coco_images}
            ann_dict = {}
            for ann in coco_annotations:
                img_id = ann["image_id"]
                if img_id not in ann_dict:
                    ann_dict[img_id] = []
                ann_dict[img_id].append(ann)

            for img_id, img in img_dict.items():
                path = image_dir.absolute() / img["file_name"]
                if not path.exists():
                    continue
                path = str(path)

                img_anns = ann_dict.get(img_id, [])
                img_h = img["height"]
                img_w = img["width"]

                for i, ann in enumerate(img_anns):
                    class_name = categories[ann["category_id"]]
                    yield {
                        "file": path,
                        "annotation": {
                            "type": "classification",
                            "class": class_name,
                        },
                    }

                    seg = ann["segmentation"]
                    if isinstance(seg, list):
                        poly = []
                        for s in seg:
                            poly_arr = np.array(s).reshape(-1, 2)
                            poly += [
                                (
                                    poly_arr[i, 0] / img_w,
                                    poly_arr[i, 1] / img_h,
                                )
                                for i in range(len(poly_arr))
                            ]
                        yield {
                            "file": path,
                            "annotation": {
                                "type": "polyline",
                                "class": class_name,
                                "points": poly,
                            },
                        }
                    else:
                        yield {
                            "file": path,
                            "annotation": {
                                "type": "rle",
                                "class": "person",
                                "height": seg["size"][0],
                                "width": seg["size"][1],
                                "counts": seg["counts"],
                            },
                        }

                    x, y, w, h = ann["bbox"]
                    yield {
                        "file": path,
                        "annotation": {
                            "type": "boundingbox",
                            "class": class_name,
                            "instance_id": i,
                            "x": x / img_w,
                            "y": y / img_h,
                            "w": w / img_w,
                            "h": h / img_h,
                        },
                    }

                    if "keypoints" in ann.keys():
                        kpts = np.array(ann["keypoints"]).reshape(-1, 3)
                        keypoints = []
                        for kp in kpts:
                            keypoints.append((kp[0] / img_w, kp[1] / img_h, int(kp[2])))
                        yield {
                            "file": path,
                            "annotation": {
                                "type": "keypoints",
                                "class": class_name,
                                "instance_id": i,
                                "keypoints": keypoints,
                            },
                        }

        added_images = self._get_added_images(generator())

        return generator(), class_names, skeletons, added_images



import builtins
import logging
import warnings
from typing import Optional

from rich import print as rprint
from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

from .environ import environ


def setup_logging(
    *,
    file: Optional[str] = None,
    use_rich: bool = False,
    level: Optional[str] = None,
    configure_warnings: bool = True,
    rich_print: bool = False,
) -> None:
    """Globally configures logging.

    Configures a standar Luxonis logger. Optionally utilizes rich library, configures
    handling of warnings (from warnings module) and saves the logs to a file.

    @type file: str or None
    @param file: Path to a file where logs will be saved. If None, logs will not be
        saved. Defaults to None.
    @type use_rich: bool
    @param use_rich: If True, rich library will be used for logging. Defaults to False.
    @type level: str or None
    @param level: Logging level. One of "DEBUG", "INFO", "WARNING", "ERROR", and
        "CRITICAL". Defaults to "INFO". The log level can be changed using "LOG_LEVEL"
        environment variable.
    @type configure_warnings: bool
    @param configure_warnings: If True, warnings will be logged. Defaults to True.
    @type rich_print: bool
    @param rich_print: If True, builtins.print will be replaced with rich.print.
        Defaults to False.
    """

    if rich_print:
        builtins.print = rprint

    # NOTE: So we can simply run e.g. `LOG_LEVEL=DEBUG python ...`
    level = level or environ.LOG_LEVEL

    if level not in ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"):
        raise ValueError(
            f"Invalid logging level: {level}. "
            "Use one of 'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'."
        )

    handlers = []

    format = file_format = "%(asctime)s [%(levelname)s] %(message)s"
    datefmt = "[%Y-%m-%d %H:%M:%S]"
    if use_rich:
        format = "%(message)s"
        # NOTE: The default rich logging colors are weird.
        _theme = Theme(
            {
                "logging.level.debug": "magenta",
                "logging.level.info": "green",
                "logging.level.warning": "yellow",
            }
        )
        _console = Console(theme=_theme)
        handlers.append(
            RichHandler(
                rich_tracebacks=True,
                tracebacks_show_locals=False,
                console=_console,
                omit_repeated_times=False,
            )
        )
    else:
        handlers.append(logging.StreamHandler())

    if file is not None:
        file_handler = logging.FileHandler(file)
        file_handler.setFormatter(logging.Formatter(file_format, datefmt=datefmt))
        handlers.append(file_handler)

    logging.basicConfig(level=level, format=format, datefmt=datefmt, handlers=handlers)

    if configure_warnings:
        logger = logging.getLogger("warnings.warn")

        def custom_warning_handler(message, *_):
            logger.warning(message)

        warnings.showwarning = custom_warning_handler


def reset_logging() -> None:
    """Resets the logging module back to its default state."""
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)



import glob
import os
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union

import numpy as np
from unique_names_generator import get_random_name


class LuxonisTracker:
    def __init__(
        self,
        project_name: Optional[str] = None,
        project_id: Optional[str] = None,
        run_name: Optional[str] = None,
        run_id: Optional[str] = None,
        save_directory: str = "output",
        is_tensorboard: bool = False,
        is_wandb: bool = False,
        is_mlflow: bool = False,
        is_sweep: bool = False,
        wandb_entity: Optional[str] = None,
        mlflow_tracking_uri: Optional[str] = None,
        rank: int = 0,
    ):
        """Implementation of PytorchLightning Logger that wraps various logging
        software. Supported loggers: TensorBoard, WandB and MLFlow.

        @type project_name: Optional[str]
        @param project_name: Name of the project used for WandB and MLFlow.
            Defaults to None.

        @type project_id: Optional[str]
        @param project_id: Project id used for WandB and MLFlow.
            Defaults to None.

        @type run_name: Optional[str]
        @param run_name: Name of the run, if None then auto-generate random name.
            Defaults to None.

        @type run_id: Optional[str]
        @param run_id: Run id used for continuing MLFlow run.
            Defaults to None.

        @type save_directory: str
        @param save_directory: Path to save directory.
            Defaults to "output".

        @type is_tensorboard: bool
        @param is_tensorboard: Wheter use TensorBoard logging.
            Defaults to False.

        @type is_wandb: bool
        @param is_wandb: Wheter use WandB logging.
            Defaults to False.

        @type is_mlflow: bool
        @param is_mlflow: Wheter use MLFlow logging.
            Defaults to False.

        @type is_sweep: bool
        @param is_sweep: Wheter current run is part of a sweep.
            Defaults to False.

        @type wandb_entity: Optional[str]
        @param wandb_entity: WandB entity to use.
            Defaults to None.

        @type mlflow_tracking_uri: Optional[str]
        @param mlflow_tracking_uri: MLFlow tracking uri to use.
            Defaults to None.

        @type rank: int
        @param rank: Rank of the process, used when running on multiple threads.
            Defaults to 0.
        """
        self.project_name = project_name
        self.project_id = project_id
        self.save_directory = save_directory
        self.is_tensorboard = is_tensorboard
        self.is_wandb = is_wandb
        self.is_mlflow = is_mlflow
        self.is_sweep = is_sweep
        self.rank = rank

        self.run_id = run_id  # if using MLFlow then it will continue previous run

        if is_wandb or is_mlflow:
            if self.project_name is None and self.project_id is None:
                raise Exception("Either project_name or project_id must be specified!")

        if self.is_wandb and wandb_entity is None:
            raise Exception("Must specify wandb_entity when using wandb!")
        else:
            self.wandb_entity = wandb_entity
        if self.is_mlflow and mlflow_tracking_uri is None:
            raise Exception("Must specify mlflow_tracking_uri when using mlflow!")
        else:
            self.mlflow_tracking_uri = mlflow_tracking_uri

        if not (self.is_tensorboard or self.is_wandb or self.is_mlflow):
            raise Exception("At least one integration must be used!")

        self._experiment = None

        if run_name:
            self.run_name = run_name
        else:
            # create new directory if rank==0 else return newest run
            if rank == 0:
                self.run_name = self._get_run_name()
            else:
                self.run_name = self._get_latest_run_name()

        Path(f"{self.save_directory}/{self.run_name}").mkdir(
            parents=True, exist_ok=True
        )

    @staticmethod
    def rank_zero_only(fn: Callable) -> Callable:
        """Function wrapper that lets only processes with rank=0 execute it."""

        @wraps(fn)
        def wrapped_fn(self, *args: Any, **kwargs: Any) -> Optional[Any]:
            if self.rank == 0:
                return fn(self, *args, **kwargs)
            return None

        return wrapped_fn

    @property
    def name(self) -> str:
        """Returns run name.

        @type: str
        """
        return self.run_name

    @property
    def version(self) -> int:
        """Returns tracker's version.

        @type: int
        """
        return 1

    @property
    @rank_zero_only
    def experiment(self) -> Dict[str, Any]:
        """Creates new experiments or returns active ones if already created."""
        if self._experiment is not None:
            return self._experiment

        self._experiment = {}

        if self.is_tensorboard:
            from torch.utils.tensorboard import SummaryWriter

            log_dir = f"{self.save_directory}/tensorboard_logs/{self.run_name}"
            self._experiment["tensorboard"] = SummaryWriter(log_dir=log_dir)

        if self.is_wandb:
            import wandb

            self._experiment["wandb"] = wandb

            log_dir = f"{self.save_directory}/wandb_logs"
            Path(log_dir).mkdir(parents=True, exist_ok=True)

            self._experiment["wandb"].init(
                project=self.project_name
                if self.project_name is not None
                else self.project_id,
                entity=self.wandb_entity,
                dir=log_dir,
                name=self.run_name,
            )

        if self.is_mlflow:
            import mlflow

            self._experiment["mlflow"] = mlflow

            self.artifacts_dir = f"{self.save_directory}/{self.run_name}/artifacts"
            Path(self.artifacts_dir).mkdir(parents=True, exist_ok=True)

            self._experiment["mlflow"].set_tracking_uri(self.mlflow_tracking_uri)

            if self.project_id is not None:
                self.project_name = None
            experiment = self._experiment["mlflow"].set_experiment(
                experiment_name=self.project_name, experiment_id=self.project_id
            )
            self.project_id = experiment.experiment_id

            # if self.run_id == None then create new run, else use alredy created one
            run = self._experiment["mlflow"].start_run(
                run_id=self.run_id, run_name=self.run_name, nested=self.is_sweep
            )
            self.run_id = run.info.run_id

        return self._experiment

    @rank_zero_only
    def log_hyperparams(
        self, params: Dict[str, Union[str, bool, int, float, None]]
    ) -> None:
        """Logs hyperparameter dictionary.

        @type params: Dict[str, Union[str, bool, int, float, None]]
        @param params: Dict of hyperparameters key-value pairs.
        """
        if self.is_tensorboard:
            self.experiment["tensorboard"].add_hparams(
                params,
                {
                    "placeholder_metric": 0
                },  # placeholder metric is needed due to this issue: https://github.com/tensorflow/tensorboard/issues/5476
            )
        if self.is_wandb:
            self.experiment["wandb"].config.update(params)
        if self.is_mlflow:
            self.experiment["mlflow"].log_params(params)

    @rank_zero_only
    def log_metric(self, name: str, value: float, step: int) -> None:
        """Logs metric value with name and step.

        @note: step is ommited when logging with wandb to avoid problems with
            inconsistent incrementation.
        @type name: str
        @param name: Metric name
        @type value: float
        @param value: Metric value
        @type step: int
        @param step: Current step
        """
        if self.is_tensorboard:
            self.experiment["tensorboard"].add_scalar(name, value, step)

        if self.is_wandb:
            # let wandb increment step to avoid calls with inconsistent steps
            self.experiment["wandb"].log({name: value})

        if self.is_mlflow:
            self.experiment["mlflow"].log_metric(name, value, step)

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, float], step: int) -> None:
        """Logs metric dictionary.

        @type metrics: Dict[str, float]
        @param metrics: Dict of metric key-value pairs
        @type step: int
        @param step: Current step
        """
        if self.is_tensorboard:
            for key, value in metrics.items():
                self.experiment["tensorboard"].add_scalar(key, value, step)
        if self.is_wandb:
            self.experiment["wandb"].log(metrics)
        if self.is_mlflow:
            self.experiment["mlflow"].log_metrics(metrics, step)

    @rank_zero_only
    def log_image(self, name: str, img: np.ndarray, step: int) -> None:
        """Logs image with name and step. Note: step is omitted when logging with wandb
        is used to avoid problems with inconsistent incrementation.

        @type name: str
        @param name: Caption of the image
        @type img: np.ndarray
        @param img: Image data
        @type step: int
        @param step: Current step
        """
        if self.is_tensorboard:
            self.experiment["tensorboard"].add_image(name, img, step, dataformats="HWC")

        if self.is_wandb:
            wandb_image = self._experiment["wandb"].Image(img, caption=name)
            # if step is added here it doesn't work correctly with wandb
            self._experiment["wandb"].log({name: wandb_image})

        if self.is_mlflow:
            # split images into separate directories based on step
            base_path, img_caption = name.rsplit("/", 1)
            self._experiment["mlflow"].log_image(
                img, f"{base_path}/{step}/{img_caption}.png"
            )

    @rank_zero_only
    def log_images(self, imgs: Dict[str, np.ndarray], step: int) -> None:
        """Logs multiple images.

        @type imgs: Dict[str, np.ndarray]
        @param imgs: Dict of image key-value pairs where key is image caption and value
            is image data
        @type step: int
        @param step: Current step
        """
        for caption, img in imgs.items():
            self.log_image(caption, img, step)

    def _get_next_run_number(self) -> int:
        """Returns number id for next run."""

        log_dirs = glob.glob(f"{self.save_directory}/*")
        log_dirs = [path.split("/")[-1] for path in log_dirs if os.path.isdir(path)]

        nums = [path.split("-")[0] for path in log_dirs]
        nums = [int(num) for num in nums if num.isnumeric()]

        if len(nums) == 0:
            return 0
        else:
            return max(nums) + 1

    def _get_run_name(self) -> str:
        """Generates new run name."""
        name_without_number = get_random_name(separator="-", style="lowercase")
        number = self._get_next_run_number()
        return f"{number}-{name_without_number}"

    def _get_latest_run_name(self) -> str:
        """Returns most recently created run name."""
        log_dirs = glob.glob(f"{self.save_directory}/*")
        log_dirs = [path for path in log_dirs if os.path.isdir(path)]
        runs = []
        for ld in log_dirs:
            ld = ld.replace(f"{self.save_directory}/", "")
            if ld.split("-")[0].isnumeric():
                runs.append(ld)
        runs.sort(
            key=lambda x: os.path.getmtime(os.path.join(self.save_directory, x)),
            reverse=True,
        )
        return runs[0]



from ..guard_extras import guard_missing_extra

with guard_missing_extra("utils"):
    from .config import LuxonisConfig
    from .environ import Environ, environ
    from .filesystem import PUT_FILE_REGISTRY, LuxonisFileSystem
    from .logging import reset_logging, setup_logging
    from .pydantic_utils import BaseModelExtraForbid
    from .registry import AutoRegisterMeta, Registry


__all__ = [
    "LuxonisConfig",
    "LuxonisFileSystem",
    "PUT_FILE_REGISTRY",
    "AutoRegisterMeta",
    "Registry",
    "setup_logging",
    "reset_logging",
    "environ",
    "Environ",
    "BaseModelExtraForbid",
]



import logging
import zipfile
from enum import Enum
from pathlib import Path
from typing import Dict, Literal, Optional, Tuple, Type, Union

from luxonis_ml.data import DATASETS_REGISTRY, BaseDataset, LuxonisDataset
from luxonis_ml.enums import DatasetType
from luxonis_ml.utils import LuxonisFileSystem

from .base_parser import BaseParser
from .classification_directory_parser import ClassificationDirectoryParser
from .coco_parser import COCOParser
from .create_ml_parser import CreateMLParser
from .darknet_parser import DarknetParser
from .segmentation_mask_directory_parser import SegmentationMaskDirectoryParser
from .solo_parser import SOLOParser
from .tensorflow_csv_parser import TensorflowCSVParser
from .voc_parser import VOCParser
from .yolov4_parser import YoloV4Parser
from .yolov6_parser import YoloV6Parser

logger = logging.getLogger(__name__)


class ParserType(Enum):
    DIR = "dir"
    SPLIT = "split"


class LuxonisParser:
    parsers: Dict[DatasetType, Type[BaseParser]] = {
        DatasetType.COCO: COCOParser,
        DatasetType.VOC: VOCParser,
        DatasetType.DARKNET: DarknetParser,
        DatasetType.YOLOV6: YoloV6Parser,
        DatasetType.YOLOV4: YoloV4Parser,
        DatasetType.CREATEML: CreateMLParser,
        DatasetType.TFCSV: TensorflowCSVParser,
        DatasetType.CLSDIR: ClassificationDirectoryParser,
        DatasetType.SEGMASK: SegmentationMaskDirectoryParser,
        DatasetType.SOLO: SOLOParser,
    }

    def __init__(
        self,
        dataset_dir: str,
        *,
        dataset_name: Optional[str] = None,
        delete_existing: bool = False,
        save_dir: Optional[Union[Path, str]] = None,
        dataset_plugin: Optional[str] = None,
        dataset_type: Optional[DatasetType] = None,
        **kwargs,
    ):
        """High-level abstraction over various parsers.

        Automatically recognizes the dataset format and uses the appropriate parser.

        @type dataset_dir: str
        @param dataset_dir: Path to the dataset directory or zip file. Can also be a
            remote URL supported by L{LuxonisFileSystem}.
        @type dataset_name: Optional[str]
        @param dataset_name: Name of the dataset. If C{None}, the name is derived from
            the name of the dataset directory.
        @type delete_existing: bool
        @param delete_existing: If existing dataset with the same name should be deleted
            before parsing.
        @type save_dir: Optional[Union[Path, str]]
        @param save_dir: If a remote URL is provided in C{dataset_dir}, the dataset will
            be downloaded to this directory. If C{None}, the dataset will be downloaded
            to the current working directory.
        @type dataset_plugin: Optional[str]
        @param dataset_plugin: Name of the dataset plugin to use. If C{None},
            C{LuxonisDataset} is used.
        @type dataset_type: Optional[DatasetType]
        @param dataset_type: If provided, the parser will use this dataset type instead
            of trying to recognize it automatically.
        @type kwargs: Dict[str, Any]
        @param kwargs: Additional C{kwargs} to be passed to the constructor of specific
            L{BaseDataset} implementation.
        """
        save_dir = Path(save_dir) if save_dir else None
        name = Path(dataset_dir).name
        local_path = (save_dir or Path.cwd()) / name
        self.dataset_dir = LuxonisFileSystem.download(dataset_dir, local_path)
        if self.dataset_dir.suffix == ".zip":
            with zipfile.ZipFile(self.dataset_dir, "r") as zip_ref:
                unzip_dir = self.dataset_dir.parent / self.dataset_dir.stem
                logger.info(f"Extracting '{self.dataset_dir.name}' to '{unzip_dir}'")
                zip_ref.extractall(unzip_dir)
                self.dataset_dir = unzip_dir

        if dataset_type:
            self.dataset_type = dataset_type
            self.parser_type = (
                ParserType.DIR
                if Path(self.dataset_dir / "train").exists()
                else ParserType.SPLIT
            )
        else:
            self.dataset_type, self.parser_type = self._recognize_dataset()

        if dataset_plugin:
            self.dataset_constructor: Type[BaseDataset] = DATASETS_REGISTRY.get(
                dataset_plugin
            )
        else:
            self.dataset_constructor = LuxonisDataset

        dataset_name = dataset_name or name.replace(" ", "_").split(".")[0]
        self.dataset_exists = self.dataset_constructor.exists(dataset_name=dataset_name)
        if delete_existing and self.dataset_exists:
            logger.warning(f"Deleting existing dataset '{dataset_name}'")
            self.dataset = self.dataset_constructor(dataset_name=dataset_name, **kwargs)
            self.dataset.delete_dataset()
            self.dataset_exists = False

        self.dataset = self.dataset_constructor(dataset_name=dataset_name, **kwargs)
        self.parser = self.parsers[self.dataset_type](self.dataset)

    def parse(self, **kwargs) -> BaseDataset:
        """Parses the dataset and returns it in LuxonisDataset format.

        If the dataset already exists, parsing will be skipped and the existing dataset
        will be returned instead.

        @type kwargs: Dict[str, Any]
        @param kwargs: Additional C{kwargs} for specific parser implementation.
        @rtype: LuxonisDataset
        @return: Parsed dataset in L{LuxonisDataset} format.
        """
        if self.dataset_exists:
            logger.warning(
                "There already exists an LDF dataset with this name. "
                "Parsing will be skipped and existing dataset will be used instead."
            )
            return self.dataset
        if self.parser_type == ParserType.DIR:
            dataset = self._parse_dir(**kwargs)
        else:
            dataset = self._parse_split(**kwargs)

        logger.info("Dataset parsed successfully.")
        return dataset

    def _recognize_dataset(self) -> Tuple[DatasetType, ParserType]:
        """Recognizes the dataset format and parser type.

        @rtype: Tuple[DatasetType, ParserType]
        @return: Tuple of dataset type and parser type.
        """
        for typ, parser in self.parsers.items():
            if parser.validate(self.dataset_dir):
                logger.info(
                    f"[cyan]Recognized dataset format as [red]<{typ.name}>",
                    extra={"markup": True},
                )
                return typ, ParserType.DIR

        for typ, parser in self.parsers.items():
            if parser.validate_split(self.dataset_dir):
                logger.info(
                    f"[cyan]Recognized dataset format as a split of [red]<{typ.name}>",
                    extra={"markup": True},
                )
                return typ, ParserType.SPLIT

        raise ValueError(
            f"Dataset {self.dataset_dir} is not in expected format for any of the parsers."
        )

    def _parse_dir(self, **kwargs) -> BaseDataset:
        """Parses all present data in LuxonisDataset format.

        Check under each parser for the expected directory structure.

        @type kwargs: Dict[str, Any]
        @param kwargs: Additional C{kwargs} for specific parser function.
        @rtype: LuxonisDataset
        @return: C{LDF} with all the images and annotations parsed.
        """

        return self.parser.parse_dir(self.dataset_dir, **kwargs)

    def _parse_split(
        self,
        split: Optional[Literal["train", "val", "test"]] = None,
        random_split: bool = True,
        split_ratios: Tuple[float, float, float] = (0.8, 0.1, 0.1),
        **kwargs,
    ) -> BaseDataset:
        """Parses data from a subdirectory representing a single split.

        Should be used if adding/changing only specific split. Check under each parser
        for expected directory structure.

        @type split: Optional[Literal["train", "val", "test"]]
        @param split: As what split the data will be added to LDF. If set,
            C{split_ratios} and C{random_split} are ignored.
        @type random_split: bool
        @param random_split: If random splits should be made. If C{True},
            C{split_ratios} are used.
        @type split_ratios: Optional[Tuple[float, float, float]]
        @param split_ratios: Ratios for random splits. Only used if C{random_split} is
            C{True}. Defaults to C{(0.8, 0.1, 0.1)}.
        @type kwargs: Dict[str, Any]
        @param kwargs: Additional kwargs for specific parser implementation.
        @rtype: LuxonisDataset
        @return: C{LDF} with all the images and annotations parsed.
        """
        parsed_kwargs = self.parser.validate_split(self.dataset_dir)
        if parsed_kwargs is None:
            raise ValueError(
                f"Dataset {self.dataset_dir} is not in the expected format for {self.dataset_type} parser."
            )
        return self.parser.parse_split(
            split, random_split, split_ratios, **parsed_kwargs, **kwargs
        )



import json
import logging
import os
import shutil
import time
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import numpy as np
import polars as pl
import pyarrow.parquet as pq
import rich.progress
from typing_extensions import Self

import luxonis_ml.data.utils.data_utils as data_utils
from luxonis_ml.utils import LuxonisFileSystem, environ
from luxonis_ml.utils.filesystem import PathType

from ..utils.constants import LDF_VERSION
from ..utils.enums import BucketStorage, BucketType
from ..utils.parquet import ParquetFileManager
from .annotation import Annotation, ArrayAnnotation, DatasetRecord
from .base_dataset import BaseDataset, DatasetIterator
from .source import LuxonisSource


class LuxonisDataset(BaseDataset):
    def __init__(
        self,
        dataset_name: Optional[str] = None,
        dataset_id: Optional[str] = None,
        team_id: Optional[str] = None,
        team_name: Optional[str] = None,
        bucket_type: BucketType = BucketType.INTERNAL,
        bucket_storage: BucketStorage = BucketStorage.LOCAL,
    ) -> None:
        """Luxonis Dataset Format (LDF) is used to define datasets in the Luxonis MLOps
        ecosystem.

        @type dataset_name: Optional[str]
        @param dataset_name: Name of the dataset
        @type team_id: Optional[str]
        @param team_id: Optional unique team identifier for the cloud
        @type team_name: Optional[str]
        @param team_name: Optional team name for the cloud
        @type dataset_id: Optional[str]
        @param dataset_id: Optional dataset ID unique identifier
        @type bucket_type: BucketType
        @param bucket_type: Whether to use external cloud buckets
        @type bucket_storage: BucketStorage
        @param bucket_storage: Underlying bucket storage from local, S3, or GCS
        @raise ValueError: If C{dataset_name} is not a string or if neither
            C{dataset_name} nor C{dataset_id} are provided.
        """
        if dataset_name is None and dataset_id is None:
            raise ValueError(
                "Must provide either dataset_name or dataset_id when initializing LuxonisDataset"
            )

        self.base_path = environ.LUXONISML_BASE_PATH
        self.base_path.mkdir(exist_ok=True)

        credentials_cache_file = self.base_path / "credentials.json"
        if credentials_cache_file.exists():
            with open(credentials_cache_file) as file:
                self.config = json.load(file)
        else:
            self.config = {}

        team_id = team_id or self._get_config("LUXONISML_TEAM_ID")
        team_name = team_name or self._get_config("LUXONISML_TEAM_NAME")

        self.bucket = self._get_config("LUXONISML_BUCKET")

        self.dataset_name = dataset_name
        self.dataset_id = dataset_id
        self.team_id = team_id
        self.team_name = team_name
        self.bucket_type = bucket_type
        self.bucket_storage = bucket_storage

        if self.bucket_storage != BucketStorage.LOCAL and self.bucket is None:
            raise Exception("Must set LUXONISML_BUCKET environment variable!")

        self.datasets_cache_file = self.base_path / "datasets.json"
        if self.datasets_cache_file.exists():
            with open(self.datasets_cache_file) as file:
                self.datasets = json.load(file)
        else:
            self.datasets = {}

        if self.dataset_name is None:
            raise Exception("Must provide a dataset_name for offline mode")
        if self.dataset_name in self.datasets:
            self.dataset_doc = self.datasets[self.dataset_name]

        else:
            self.source = LuxonisSource("default")
            self.datasets[self.dataset_name] = {
                "source": self.source.to_document(),
                "ldf_version": LDF_VERSION,
                "classes": {},
            }
            self._write_datasets()

        self._init_path()

        if self.bucket_storage == BucketStorage.LOCAL:
            self.fs = LuxonisFileSystem(f"file://{self.path}")
        elif self.bucket_storage == BucketStorage.AZURE_BLOB:
            raise NotImplementedError
        else:
            self.tmp_dir = environ.LUXONISML_TMP_DIR
            self.fs = LuxonisFileSystem(self.path)

        self.logger = logging.getLogger(__name__)

        self.progress = rich.progress.Progress(
            rich.progress.TextColumn("[progress.description]{task.description}"),
            rich.progress.BarColumn(),
            rich.progress.TaskProgressColumn(),
            rich.progress.MofNCompleteColumn(),
            rich.progress.TimeRemainingColumn(),
        )

    @classmethod
    def delete_and_create(
        cls,
        name: str,
        remote: bool = False,
        bucket_storage: BucketStorage = BucketStorage.LOCAL,
        bucket: Optional[str] = None,
        team_id: Optional[str] = None,
        **kwargs,
    ) -> "LuxonisDataset":
        if remote and bucket_storage == BucketStorage.LOCAL:
            raise ValueError("Cannot create a remote dataset with local storage")
        if LuxonisDataset.exists(
            name,
            remote=remote,
            bucket_storage=bucket_storage,
            bucket=bucket,
            team_id=team_id,
        ):
            LuxonisDataset(
                name, bucket_storage=bucket_storage, team_id=team_id, **kwargs
            ).delete_dataset(delete_remote=remote)
        return LuxonisDataset(
            name, bucket_storage=bucket_storage, team_id=team_id, **kwargs
        )

    @property
    def identifier(self) -> str:
        if self.dataset_name is not None:
            return self.dataset_name
        assert (
            self.dataset_id is not None
        ), "At least one of dataset_name or dataset_id must be provided."
        return self.dataset_id

    def __len__(self) -> int:
        """Returns the number of instances in the dataset."""

        df = self._load_df_offline(self.bucket_storage != BucketStorage.LOCAL)
        return len(df.select("uuid").unique()) if df is not None else 0

    def _write_datasets(self) -> None:
        with open(self.datasets_cache_file, "w") as file:
            json.dump(self.datasets, file, indent=4)

    def _get_config(self, key: str) -> str:
        """Gets secret credentials from credentials file or ENV variables."""

        if key in self.config.keys():
            return self.config[key]
        else:
            if not hasattr(environ, key):
                raise Exception(f"Must set {key} in ENV variables")
            return getattr(environ, key)

    def _init_path(self) -> None:
        """Configures local path or bucket directory."""

        self.local_path = (
            self.base_path / "data" / self.team_id / "datasets" / self.identifier
        )
        self.media_path = self.local_path / "media"
        self.annotations_path = self.local_path / "annotations"
        self.metadata_path = self.local_path / "metadata"
        self.masks_path = self.local_path / "masks"

        if self.bucket_storage == BucketStorage.LOCAL:
            self.path = str(self.local_path)
            for path in [
                self.media_path,
                self.annotations_path,
                self.metadata_path,
            ]:
                path.mkdir(exist_ok=True, parents=True)
        else:
            assert self.dataset_name
            self.path = self._construct_url(
                self.bucket_storage, self.bucket, self.team_id, self.dataset_name
            )

    @staticmethod
    def _construct_url(
        bucket_storage: BucketStorage, bucket: str, team_id: str, dataset_name: str
    ) -> str:
        """Constructs a URL for a remote dataset."""
        return f"{bucket_storage.value}://{bucket}/{team_id}/datasets/{dataset_name}"

    def _load_df_offline(self, sync_mode: bool = False) -> Optional[pl.DataFrame]:
        dfs = []
        if self.bucket_storage == BucketStorage.LOCAL or sync_mode:
            annotations_path = self.annotations_path
        else:
            annotations_path = self.tmp_dir / "annotations"
        if not annotations_path.exists():
            return None
        for file in annotations_path.iterdir():
            if file.suffix == ".parquet":
                dfs.append(pl.read_parquet(file))
        if dfs:
            return pl.concat(dfs)
        else:
            return None

    def _find_filepath_uuid(
        self,
        filepath: Path,
        index: Optional[pl.DataFrame],
        *,
        raise_on_missing: bool = False,
    ) -> Optional[str]:
        if index is None:
            return None

        abs_path = str(filepath.absolute())
        matched = index.filter(pl.col("original_filepath") == abs_path)

        if len(matched):
            return list(matched.select("uuid"))[0][0]
        elif raise_on_missing:
            raise ValueError(f"File {abs_path} not found in index")
        return None

    def _get_file_index(self) -> Optional[pl.DataFrame]:
        index = None
        if self.bucket_storage == BucketStorage.LOCAL:
            file_index_path = self.metadata_path / "file_index.parquet"
        else:
            file_index_path = self.tmp_dir / "file_index.parquet"
            try:
                self.fs.get_file("metadata/file_index.parquet", file_index_path)
            except Exception:
                pass
        if file_index_path.exists():
            index = pl.read_parquet(file_index_path).select(
                pl.all().exclude("^__index_level_.*$")
            )
        return index

    def _write_index(
        self,
        index: Optional[pl.DataFrame],
        new_index: Dict[str, List[str]],
        override_path: Optional[str] = None,
    ) -> None:
        if override_path:
            file_index_path = override_path
        else:
            file_index_path = self.metadata_path / "file_index.parquet"
        df = pl.DataFrame(new_index)
        if index is not None:
            df = pl.concat([index, df])
        pq.write_table(df.to_arrow(), file_index_path)

    @contextmanager
    def _log_time(self):
        t = time.time()
        yield
        self.logger.info(f"Took {time.time() - t} seconds")

    def _make_temp_dir(self, remove_previous: bool = True) -> None:
        if self.tmp_dir.exists() and remove_previous:
            self._remove_temp_dir()
        self.tmp_dir.mkdir(exist_ok=True)

    def _remove_temp_dir(self) -> None:
        shutil.rmtree(self.tmp_dir)

    def update_source(self, source: LuxonisSource) -> None:
        """Updates underlying source of the dataset with a new L{LuxonisSource}.

        @type source: L{LuxonisSource}
        @param source: The new L{LuxonisSource} to replace the old one.
        """

        self.datasets[self.dataset_name]["source"] = source.to_document()
        self._write_datasets()
        self.source = source

    def _write_classes(self, *, _remove_tmp_dir: bool) -> None:
        self._write_datasets()

        if self.bucket_storage != BucketStorage.LOCAL:
            classes = self.datasets[self.dataset_name]["classes"]
            self._make_temp_dir(remove_previous=_remove_tmp_dir)
            local_file = self.tmp_dir / "classes.json"
            with open(local_file, "w") as file:
                json.dump(classes, file, indent=4)
            self.fs.put_file(local_file, "metadata/classes.json")
            if _remove_tmp_dir:
                self._remove_temp_dir()

    def _read_classes(self, sync_mode: bool) -> Dict[str, Any]:
        if sync_mode:
            local_file = self.metadata_path / "classes.json"
            self.fs.get_file("metadata/classes.json", local_file)
            with open(local_file) as file:
                return json.load(file)

        return self.datasets[self.dataset_name]["classes"]

    def set_classes(
        self,
        classes: List[str],
        task: Optional[str] = None,
        *,
        _remove_tmp_dir: bool = True,
    ) -> None:
        if task is not None:
            self.datasets[self.dataset_name]["classes"][task] = classes
        else:
            raise NotImplementedError(
                "Setting classes for all tasks not yet supported. "
                "Set classes individually for each task"
            )
        self._write_classes(_remove_tmp_dir=_remove_tmp_dir)

    def get_classes(
        self, sync_mode: bool = False
    ) -> Tuple[List[str], Dict[str, List[str]]]:
        classes = set()
        classes_by_task = {}
        classes_json = self._read_classes(sync_mode)
        for task in classes_json:
            task_classes = classes_json[task]
            if len(task_classes):
                classes_by_task[task] = task_classes
                for cls_ in task_classes:
                    classes.add(cls_)
        classes = sorted(list(classes))

        return classes, classes_by_task

    def set_skeletons(
        self, skeletons: Dict[str, Dict], task: Optional[str] = None
    ) -> None:
        if task is None:
            raise NotImplementedError("Skeletons must be set for a specific task")

        if "skeletons" not in self.datasets[self.dataset_name]:
            self.datasets[self.dataset_name]["skeletons"] = {}
        self.datasets[self.dataset_name]["skeletons"][task] = skeletons
        self._write_datasets()

    def get_skeletons(self) -> Dict[str, Dict]:
        if "skeletons" not in self.datasets[self.dataset_name]:
            self.datasets[self.dataset_name]["skeletons"] = {}
        return self.datasets[self.dataset_name]["skeletons"]

    def sync_from_cloud(self) -> None:
        """Downloads data from a remote cloud bucket."""

        if self.bucket_storage == BucketStorage.LOCAL:
            self.logger.warning("This is a local dataset! Cannot sync")
        else:
            if not getattr(self, "is_synced", False):
                local_dir = self.base_path / "data" / self.team_id / "datasets"
                if not local_dir.exists():
                    os.makedirs(local_dir, exist_ok=True)

                self.fs.get_dir(remote_paths="", local_dir=local_dir)

                self.is_synced = True

    def delete_dataset(self, *, delete_remote: bool = False) -> None:
        """Deletes the dataset from local storage and optionally from the cloud.

        @type delete_remote: bool
        @param delete_remote: Whether to delete the dataset from the cloud.
        """
        del self.datasets[self.dataset_name]
        self._write_datasets()
        self.logger.info(f"Deleted dataset {self.dataset_name}")
        if self.bucket_storage == BucketStorage.LOCAL:
            shutil.rmtree(self.path)
        elif delete_remote:
            self.logger.info(f"Deleting dataset {self.dataset_name} from cloud")
            assert self.path
            assert self.dataset_name
            self.fs.delete_dir(allow_delete_parent=True)

    def _process_arrays(self, batch_data: List[DatasetRecord]) -> None:
        array_paths = set(
            ann.path for ann in batch_data if isinstance(ann, ArrayAnnotation)
        )
        if array_paths:
            task = self.progress.add_task(
                "[magenta]Processing arrays...", total=len(batch_data)
            )
            self.logger.info("Checking arrays...")
            with self._log_time():
                data_utils.check_arrays(array_paths)
            self.logger.info("Generating array UUIDs...")
            with self._log_time():
                array_uuid_dict = self.fs.get_file_uuids(
                    array_paths, local=True
                )  # TODO: support from bucket
            if self.bucket_storage != BucketStorage.LOCAL:
                self.logger.info("Uploading arrays...")
                # TODO: support from bucket (likely with a self.fs.copy_dir)
                with self._log_time():
                    arrays_upload_dict = self.fs.put_dir(
                        local_paths=array_paths,
                        remote_dir="arrays",
                        uuid_dict=array_uuid_dict,
                    )
            self.logger.info("Finalizing paths...")
            self.progress.start()
            for ann in batch_data:
                if isinstance(ann, ArrayAnnotation):
                    if self.bucket_storage != BucketStorage.LOCAL:
                        remote_path = arrays_upload_dict[str(ann.path)]  # type: ignore
                        remote_path = (
                            f"{self.fs.protocol}://{self.fs.path / remote_path}"
                        )
                        ann.path = remote_path  # type: ignore
                    else:
                        ann.path = ann.path.absolute()
                    self.progress.update(task, advance=1)
            self.progress.stop()
            self.progress.remove_task(task)

    def _add_process_batch(
        self,
        batch_data: List[DatasetRecord],
        pfm: ParquetFileManager,
        index: Optional[pl.DataFrame],
        new_index: Dict[str, List[str]],
        processed_uuids: Set[str],
    ) -> None:
        paths = list(set(data.file for data in batch_data))
        self.logger.info("Generating UUIDs...")
        with self._log_time():
            uuid_dict = self.fs.get_file_uuids(
                paths, local=True
            )  # TODO: support from bucket
        if self.bucket_storage != BucketStorage.LOCAL:
            self.logger.info("Uploading media...")
            # TODO: support from bucket (likely with a self.fs.copy_dir)

            with self._log_time():
                self.fs.put_dir(
                    local_paths=paths, remote_dir="media", uuid_dict=uuid_dict
                )

        task = self.progress.add_task(
            "[magenta]Processing data...", total=len(batch_data)
        )

        self._process_arrays(batch_data)

        self.logger.info("Saving annotations...")
        with self._log_time():
            self.progress.start()
            for ann in batch_data:
                filepath = ann.file
                file = filepath.name
                uuid = uuid_dict[str(filepath)]
                matched_id = self._find_filepath_uuid(filepath, index)
                if matched_id is not None:
                    if matched_id != uuid:
                        # TODO: not sure if this should be an exception or how we should really handle it
                        raise ValueError(
                            f"{filepath} already added to the dataset! Please skip or rename the file."
                        )
                        # TODO: we may also want to check for duplicate uuids to get a one-to-one relationship
                elif uuid not in processed_uuids:
                    new_index["uuid"].append(uuid)
                    new_index["file"].append(file)
                    new_index["original_filepath"].append(str(filepath.absolute()))
                    processed_uuids.add(uuid)

                pfm.write({"uuid": uuid, **ann.to_parquet_dict()})
                self.progress.update(task, advance=1)
            self.progress.stop()
            self.progress.remove_task(task)

    def _infer_task(self, ann: Annotation) -> str:
        if not hasattr(LuxonisDataset._infer_task, "_logged_inferred_classes"):
            LuxonisDataset._infer_task._logged_inferred_classes = defaultdict(bool)

        def _log_once(cls_: str, message: str, level: str = "info"):
            if not LuxonisDataset._infer_task._logged_inferred_classes[cls_]:
                LuxonisDataset._infer_task._logged_inferred_classes[cls_] = True
                getattr(self.logger, level)(message, extra={"markup": True})

        cls_ = ann.class_
        _, current_classes = self.get_classes()
        inferred_task = None

        for task, classes in current_classes.items():
            if cls_ in classes:
                if inferred_task is not None:
                    _log_once(
                        cls_,
                        f"Class [red italic]{cls_}[reset] is ambiguous between tasks [magenta italic]{inferred_task}[reset] and [magenta italic]{task}[reset]. Task inference failed.",
                        "warning",
                    )
                    inferred_task = None
                    break
                inferred_task = task
        if inferred_task is None:
            _log_once(
                cls_,
                f"Task inference for class [red italic]{cls_}[reset] failed. "
                f"Autogenerated task [magenta italic]{ann.task}[reset] will be used.",
                "warning",
            )
        else:
            _log_once(
                cls_,
                f"Class [red italic]{cls_}[reset] inferred to belong to task [magenta italic]{inferred_task}[reset]",
            )
            return inferred_task

        return ann.task

    def add(self, generator: DatasetIterator, batch_size: int = 1_000_000) -> Self:
        if self.bucket_storage == BucketStorage.LOCAL:
            annotations_dir = self.annotations_path
        else:
            self._make_temp_dir()
            annotations_dir = self.tmp_dir / "annotations"
            annotations_dir.mkdir(exist_ok=True, parents=True)

        index = self._get_file_index()
        new_index = {"uuid": [], "file": [], "original_filepath": []}
        processed_uuids = set()

        batch_data: list[DatasetRecord] = []

        classes_per_task: Dict[str, Set[str]] = defaultdict(set)
        num_kpts_per_task: Dict[str, int] = {}

        with ParquetFileManager(annotations_dir) as pfm:
            for i, data in enumerate(generator, start=1):
                record = (
                    data if isinstance(data, DatasetRecord) else DatasetRecord(**data)
                )
                ann = record.annotation
                if ann is not None:
                    if ann.task == ann._label_type.value and (
                        isinstance(data, DatasetRecord)
                        or "task" not in data.get("annotation", {})
                    ):
                        ann.task = self._infer_task(ann)

                    classes_per_task[ann.task].add(ann.class_)
                    if ann.type_ == "keypoints":
                        num_kpts_per_task[ann.task] = len(ann.keypoints)

                batch_data.append(record)
                if i % batch_size == 0:
                    self._add_process_batch(
                        batch_data, pfm, index, new_index, processed_uuids
                    )
                    batch_data = []

            self._add_process_batch(batch_data, pfm, index, new_index, processed_uuids)

        _, curr_classes = self.get_classes()
        for task, classes in classes_per_task.items():
            old_classes = set(curr_classes.get(task, []))
            new_classes = list(classes - old_classes)
            if new_classes:
                self.logger.info(f"Detected new classes for task {task}: {new_classes}")
                self.set_classes(
                    list(classes | old_classes), task, _remove_tmp_dir=False
                )

        if self.bucket_storage == BucketStorage.LOCAL:
            self._write_index(index, new_index)
        else:
            file_index_path = str(self.tmp_dir / "file_index.parquet")
            self._write_index(index, new_index, override_path=file_index_path)
            self.fs.put_dir(Path(annotations_dir), "annotations")  # type: ignore
            self.fs.put_file(file_index_path, "metadata/file_index.parquet")
            self._remove_temp_dir()

        return self

    def make_splits(
        self,
        ratios: Tuple[float, float, float] = (0.8, 0.1, 0.1),
        definitions: Optional[Dict[str, List[PathType]]] = None,
    ) -> None:
        new_splits = {"train": [], "val": [], "test": []}
        splits_to_update = []

        if definitions is None:  # random split
            if self.bucket_storage != BucketStorage.LOCAL:
                self._make_temp_dir()
                self.fs.get_dir("annotations", self.tmp_dir / "annotations")

            df = self._load_df_offline()
            assert df is not None
            ids = df.select("uuid").unique().get_column("uuid").to_list()
            np.random.shuffle(ids)
            N = len(ids)
            b1 = round(N * ratios[0])
            b2 = round(N * ratios[0]) + round(N * ratios[1])
            new_splits["train"] = ids[:b1]
            new_splits["val"] = ids[b1:b2]
            new_splits["test"] = ids[b2:]
            splits_to_update = ["train", "val", "test"]
        else:  # provided split
            index = self._get_file_index()
            if index is None:
                raise Exception("File index not found")
            for split in "train", "val", "test":
                if split not in definitions:
                    continue
                splits_to_update.append(split)
                filepaths = definitions[split]
                if not isinstance(filepaths, list):
                    raise Exception("Must provide splits as a list of str")
                ids = [
                    self._find_filepath_uuid(
                        Path(filepath), index, raise_on_missing=True
                    )
                    for filepath in filepaths
                ]
                new_splits[split] = ids

        if self.bucket_storage == BucketStorage.LOCAL:
            splits_path = os.path.join(self.metadata_path, "splits.json")
            if os.path.exists(splits_path):
                with open(splits_path, "r") as file:
                    splits = json.load(file)
                for split in splits_to_update:
                    splits[split] = new_splits[split]
            else:
                splits = new_splits
            with open(os.path.join(self.metadata_path, "splits.json"), "w") as file:
                json.dump(splits, file, indent=4)
        else:
            remote_splits_path = "metadata/splits.json"
            local_splits_path = os.path.join(self.tmp_dir, "splits.json")
            if self.fs.exists(remote_splits_path):
                self.fs.get_file(remote_splits_path, local_splits_path)
                with open(local_splits_path, "r") as file:
                    splits = json.load(file)
                for split in splits_to_update:
                    splits[split] = new_splits[split]
            else:
                splits = new_splits
            with open(local_splits_path, "w") as file:
                json.dump(splits, file, indent=4)
            self.fs.put_file(local_splits_path, "metadata/splits.json")
            self._remove_temp_dir()

    @staticmethod
    def exists(
        dataset_name: str,
        *,
        remote: bool = False,
        bucket_storage: BucketStorage = BucketStorage.LOCAL,
        bucket: Optional[str] = None,
        team_id: Optional[str] = None,
    ) -> bool:
        """Checks if a dataset exists.

        @type dataset_name: str
        @param dataset_name: Name of the dataset to check
        @type remote: bool
        @param remote: Whether to check if the dataset exists in the cloud
        """
        if not remote:
            return dataset_name in LuxonisDataset.list_datasets()

        bucket = bucket or environ.LUXONISML_BUCKET
        if bucket is None:
            raise ValueError("Must provide a bucket name for remote `exists` check")
        team_id = team_id or environ.LUXONISML_TEAM_ID

        url = LuxonisDataset._construct_url(
            bucket_storage, bucket, team_id, dataset_name
        )
        return LuxonisFileSystem(url).exists()

    @staticmethod
    def list_datasets() -> Dict:
        """Returns a dictionary of all datasets.

        @rtype: Dict
        @return: Dictionary of all datasets
        """
        base_path = environ.LUXONISML_BASE_PATH
        datasets_cache_file = base_path / "datasets.json"
        if datasets_cache_file.exists():
            with open(datasets_cache_file) as file:
                datasets = json.load(file)
        else:
            datasets = {}

        return datasets

    def get_tasks(self) -> List[str]:
        return list(self.get_classes()[1].keys())



from typing import List, Optional

from pydantic import Field, model_validator

from luxonis_ml.utils import BaseModelExtraForbid

from ..enums import DataType, InputType


class PreprocessingBlock(BaseModelExtraForbid):
    """Represents preprocessing operations applied to the input data.

    @type mean: list
    @ivar mean: Mean values in channel order. Typically, this is BGR order.
    @type scale: list
    @ivar scale: Standardization values in channel order. Typically, this is BGR order.
    @type reverse_channels: bool | None
    @ivar reverse_channels: If True, color channels are reversed (e.g. BGR to RGB or
        vice versa).
    @type interleaved_to_planar: bool | None
    @ivar interleaved_to_planar: If True, format is changed from interleaved to planar.
    """

    mean: Optional[List[float]] = Field(
        None, description="Mean values in channel order. Typically, this is BGR order."
    )
    scale: Optional[List[float]] = Field(
        None,
        description="Standardization values in channel order. Typically, this is BGR order.",
    )
    reverse_channels: Optional[bool] = Field(
        None,
        description="If True, color channels are reversed (e.g. BGR to RGB or vice versa).",
    )
    interleaved_to_planar: Optional[bool] = Field(
        None, description="If True, format is changed from interleaved to planar."
    )


class Input(BaseModelExtraForbid):
    """Represents input stream of a model.

    @type name: str
    @ivar name: Name of the input layer.

    @type dtype: DataType
    @ivar dtype: Data type of the input data (e.g., 'float32').

    @type input_type: InputType
    @ivar input_type: Type of input data (e.g., 'image').

    @type shape: list
    @ivar shape: Shape of the input data as a list of integers (e.g. [H,W], [H,W,C], [N,H,W,C], ...).

    @type layout: str
    @ivar layout: Lettercode interpretation of the input data dimensions (e.g., 'NCHW').

    @type preprocessing: PreprocessingBlock
    @ivar preprocessing: Preprocessing steps applied to the input data.
    """

    name: str = Field(description="Name of the input layer.")
    dtype: DataType = Field(
        description="Data type of the input data (e.g., 'float32')."
    )
    input_type: InputType = Field(description="Type of input data (e.g., 'image').")
    shape: List[int] = Field(
        min_length=1,
        max_length=5,
        description="Shape of the input data as a list of integers (e.g. [H,W], [H,W,C], [N,H,W,C], ...).",
    )
    layout: str = Field(
        description="Lettercode interpretation of the input layout (e.g., 'NCHW').",
        default="NCHW",
    )
    preprocessing: PreprocessingBlock = Field(
        description="Preprocessing steps applied to the input data."
    )

    @model_validator(mode="before")
    def validate_layout(
        cls,
        values,
    ):
        if "layout" in values.keys():
            values["layout"] = values["layout"].upper()
            idx_n = values["layout"].find("N")

            if len(values["layout"]) != len(values["shape"]):
                raise ValueError("Layout and shape must have the same length.")

            if idx_n not in [-1, 0]:
                raise ValueError("N (batch size) must be the first letter if included.")

            if values["input_type"] == InputType.IMAGE.value:
                if "C" not in values["layout"]:
                    raise ValueError(
                        "C letter must be present in layout for image input type."
                    )

            if len(values["layout"]) != len(set(values["layout"])):
                raise ValueError("Layout must not contain any duplicate letters.")

        return values



from typing import Final, Set, cast

import pytest

from luxonis_ml.data import (
    Augmentations,
    BucketStorage,
    LuxonisDataset,
    LuxonisLoader,
    LuxonisParser,
)
from luxonis_ml.enums import DatasetType

SKELETONS: Final[dict] = {
    "keypoints": {
        "person": {
            "labels": [
                "nose",
                "left_eye",
                "right_eye",
                "left_ear",
                "right_ear",
                "left_shoulder",
                "right_shoulder",
                "left_elbow",
                "right_elbow",
                "left_wrist",
                "right_wrist",
                "left_hip",
                "right_hip",
                "left_knee",
                "right_knee",
                "left_ankle",
                "right_ankle",
            ],
            "edges": [
                [15, 13],
                [13, 11],
                [16, 14],
                [14, 12],
                [11, 12],
                [5, 11],
                [6, 12],
                [5, 6],
                [5, 7],
                [6, 8],
                [7, 9],
                [8, 10],
                [1, 2],
                [0, 1],
                [0, 2],
                [1, 3],
                [2, 4],
                [3, 5],
                [4, 6],
            ],
        }
    }
}
URL_PREFIX: Final[str] = "gs://luxonis-test-bucket/luxonis-ml-test-data"
WORK_DIR: Final[str] = "tests/data/parser_datasets"
DATASET_NAME: Final[str] = "__test_coco"
TASKS: Final[Set[str]] = {"segmentation", "classification", "keypoints", "boundingbox"}


@pytest.fixture(scope="function", autouse=True)
def delete_dataset():
    yield
    LuxonisDataset(DATASET_NAME).delete_dataset()


@pytest.mark.parametrize(
    ("bucket_storage",),
    [
        (BucketStorage.LOCAL,),
        (BucketStorage.S3,),
        (BucketStorage.GCS,),
    ],
)
def test_dataset(bucket_storage: BucketStorage, subtests):
    with subtests.test("test_create", bucket_storage=bucket_storage):
        parser = LuxonisParser(
            f"{URL_PREFIX}/COCO_people_subset.zip",
            dataset_name=DATASET_NAME,
            delete_existing=True,
            save_dir=WORK_DIR,
            dataset_type=DatasetType.COCO,
            bucket_storage=bucket_storage,
        )
        dataset = cast(LuxonisDataset, parser.parse())
        assert LuxonisDataset.exists(DATASET_NAME)
        assert dataset.get_classes()[0] == ["person"]
        assert set(dataset.get_tasks()) == TASKS
        assert dataset.get_skeletons() == SKELETONS

    if "dataset" not in locals():
        pytest.exit("Dataset creation failed")

    with subtests.test("test_load", bucket_storage=bucket_storage):
        loader = LuxonisLoader(dataset)
        for img, labels in loader:
            assert img is not None
            for task in TASKS:
                assert task in labels

    with subtests.test("test_load_aug", bucket_storage=bucket_storage):
        aug_config = [
            {
                "name": "Mosaic4",
                "params": {"out_width": 416, "out_height": 416, "p": 1.0},
            },
            {"name": "Defocus", "params": {"p": 1.0}},
            {"name": "Sharpen", "params": {"p": 1.0}},
            {"name": "Flip", "params": {"p": 1.0}},
            {"name": "RandomRotate90", "params": {"p": 1.0}},
        ]
        augmentations = Augmentations([512, 512], aug_config)
        loader = LuxonisLoader(dataset, augmentations=augmentations)
        for img, labels in loader:
            assert img is not None
            for task in TASKS:
                assert task in labels

    with subtests.test("test_delete", bucket_storage=bucket_storage):
        dataset.delete_dataset()
        assert not LuxonisDataset.exists(DATASET_NAME)



from typing import Literal

from pydantic import Field

from luxonis_ml.utils import BaseModelExtraForbid

from .model import Model

CONFIG_VERSION = Literal["1.0"]


class Config(BaseModelExtraForbid):
    """The main class of the multi/single-stage model config scheme (multi- stage models
    consists of interconnected single-stage models).

    @type config_version: str
    @ivar config_version: Static variable representing the version of the config scheme.
    @type model: Model
    @ivar model: A Model object representing the neural network used in the archive.
    """

    config_version: CONFIG_VERSION = Field(
        ...,
        description="Static variable representing the version of the config scheme.",
    )
    model: Model = Field(
        description="A Model object representing the neural network used in the archive."
    )



from enum import Enum


class ObjectDetectionSubtypeYOLO(Enum):
    """Object detection decoding subtypes for YOLO networks.

    Subtype members have exactly the same decoding.
    """

    YOLOv5 = "yolov5"
    YOLOv6 = "yolov6"
    YOLOv6r2 = "yolov6r2"
    YOLOv7 = "yolov7"
    YOLOv8 = "yolov8"
    YOLOv10 = "yolov10"


class ObjectDetectionSubtypeSSD(Enum):
    """Object detection decoding subtypes for SSD networks.

    Subtype members have exactly the same decoding.
    """

    SSD_PARSED = "ssd-parsed"
